from dataclasses import dataclass, asdict
from typing import Optional, Dict

from airunner.data.models import Chatbot, LLMGeneratorSettings


@dataclass
class LLMRequest:
    """
    Represents a request to a Large Language Model.

    This dataclass stores parameters for LLM generation, providing methods
    to convert between various formats and sources.

    Attributes:
        do_sample: Whether to use sampling for generation.
        early_stopping: Whether to stop generation when all beams are finished.
        eta_cutoff: Eta value cutoff for generation.
        length_penalty: Exponential penalty to the length of generated sequences.
        max_new_tokens: Maximum number of tokens to generate.
        min_length: Minimum length of the generated text.
        no_repeat_ngram_size: Size of n-grams that should not be repeated.
        num_beams: Number of beams for beam search.
        num_return_sequences: Number of sequences to return.
        repetition_penalty: Penalty for repeating tokens.
        temperature: Temperature for sampling.
        top_k: Keep only top-k tokens with highest probability.
        top_p: Keep the top tokens with cumulative probability >= top_p.
        use_cache: Whether to use the past key/values cache.
        do_tts_reply: Whether to convert the reply to speech.
    """

    do_sample: bool = True
    early_stopping: bool = True
    eta_cutoff: int = 200
    length_penalty: float = 1.0
    max_new_tokens: int = 200
    min_length: int = 1
    no_repeat_ngram_size: int = 2
    num_beams: int = 1
    num_return_sequences: int = 1
    repetition_penalty: float = 1.0
    temperature: float = 1.0
    top_k: int = 50
    top_p: float = 0.9
    use_cache: bool = True
    do_tts_reply: bool = True
    node_id: Optional[str] = None
    use_memory: bool = True

    def to_dict(self) -> Dict:
        """
        Convert the request parameters to a dictionary suitable for API calls.

        Ensures all values meet minimum thresholds and handles special cases
        for beam-based generation parameters.

        Returns:
            Dict: Dictionary representation of the request parameters.
        """
        min_val = 0.0001
        length_penalty = max(self.length_penalty, min_val)
        repetition_penalty = max(self.repetition_penalty, min_val)
        top_p = max(self.top_p, min_val)
        temperature = max(self.temperature, min_val)

        data = asdict(self)
        data.update(
            {
                "length_penalty": length_penalty,
                "repetition_penalty": repetition_penalty,
                "top_p": top_p,
                "temperature": temperature,
            }
        )

        # Length penalty flag is only used in beam-based generation modes
        # so num_beams should be > 1 for length penalty to be used.
        if self.num_beams == 1 and length_penalty != 0.0:
            del data["length_penalty"]

        # Early stopping flag is only used in beam-based generation modes
        # so num_beams should be > 1 for early stopping to be used.
        if self.num_beams == 1:
            del data["early_stopping"]

        data.pop("node_id")
        data.pop("use_memory")

        return data

    @classmethod
    def from_values(
        cls,
        do_sample: bool,
        early_stopping: bool,
        eta_cutoff: int,
        length_penalty: float,
        max_new_tokens: int,
        min_length: int,
        no_repeat_ngram_size: int,
        num_beams: int,
        num_return_sequences: int,
        repetition_penalty: float,
        temperature: float,
        top_k: int,
        top_p: float,
        use_cache: bool,
    ) -> "LLMRequest":
        """
        Create an LLMRequest instance from individual parameter values.

        Args:
            do_sample: Whether to use sampling for generation.
            early_stopping: Whether to stop generation when all beams are finished.
            eta_cutoff: Eta value cutoff for generation.
            length_penalty: Exponential penalty to the length (already scaled).
            max_new_tokens: Maximum number of tokens to generate.
            min_length: Minimum length of the generated text.
            no_repeat_ngram_size: Size of n-grams that should not be repeated.
            num_beams: Number of beams for beam search.
            num_return_sequences: Number of sequences to return.
            repetition_penalty: Penalty for repeating tokens (already scaled).
            temperature: Temperature for sampling (already scaled).
            top_k: Keep only top-k tokens with highest probability.
            top_p: Keep the top tokens with cumulative probability >= top_p (already scaled).
            use_cache: Whether to use the past key/values cache.

        Returns:
            LLMRequest: A new instance with the specified parameter values.
        """
        return cls(
            do_sample=do_sample,
            early_stopping=early_stopping,
            eta_cutoff=eta_cutoff,
            length_penalty=length_penalty / 1000.0,
            max_new_tokens=max_new_tokens,
            min_length=min_length,
            no_repeat_ngram_size=no_repeat_ngram_size,
            num_beams=num_beams,
            num_return_sequences=num_return_sequences,
            repetition_penalty=repetition_penalty / 100.0,
            temperature=temperature / 10000.0,
            top_k=top_k,  # Ensure top_k is correctly passed through
            top_p=top_p / 1000.0,
            use_cache=use_cache,
        )

    @classmethod
    def from_chatbot(cls, chatbot_id: int = None) -> "LLMRequest":
        """
        Create an LLMRequest instance from a Chatbot model.

        Args:
            chatbot_id: Optional ID of the chatbot to use. If None, uses the first chatbot.

        Returns:
            LLMRequest: A new instance with parameters from the specified chatbot.
        """
        if chatbot_id:
            chatbot = Chatbot.objects.get(chatbot_id)
        else:
            chatbot = Chatbot.objects.first()

        return cls.from_values(
            do_sample=chatbot.do_sample,
            early_stopping=chatbot.early_stopping,
            eta_cutoff=chatbot.eta_cutoff,
            length_penalty=chatbot.length_penalty,
            max_new_tokens=chatbot.max_new_tokens,
            min_length=chatbot.min_length,
            no_repeat_ngram_size=chatbot.ngram_size,
            num_beams=chatbot.num_beams,
            num_return_sequences=chatbot.num_return_sequences,
            repetition_penalty=chatbot.repetition_penalty,
            temperature=chatbot.temperature,
            top_k=chatbot.top_k,
            top_p=chatbot.top_p,
            use_cache=chatbot.use_cache,
        )

    @classmethod
    def from_llm_settings(
        cls, llm_settings_id: Optional[int] = None
    ) -> "LLMRequest":
        """
        Create an LLMRequest instance from LLMGeneratorSettings.

        Args:
            llm_settings_id: Optional ID of the settings to use. If None, uses the first settings.

        Returns:
            LLMRequest: A new instance with parameters from the specified settings or from
                        the associated chatbot if settings don't override parameters.
        """
        if llm_settings_id:
            llm_settings = LLMGeneratorSettings.objects.get(llm_settings_id)
        else:
            llm_settings = LLMGeneratorSettings.objects.first()

        if llm_settings.override_parameters:
            return cls.from_values(
                do_sample=llm_settings.do_sample,
                early_stopping=llm_settings.early_stopping,
                eta_cutoff=llm_settings.eta_cutoff,
                length_penalty=llm_settings.length_penalty,
                max_new_tokens=llm_settings.max_new_tokens,
                min_length=llm_settings.min_length,
                no_repeat_ngram_size=llm_settings.ngram_size,
                num_beams=llm_settings.num_beams,
                num_return_sequences=llm_settings.sequences,
                repetition_penalty=llm_settings.repetition_penalty,
                temperature=llm_settings.temperature,
                top_k=llm_settings.top_k,
                top_p=llm_settings.top_p,
                use_cache=llm_settings.use_cache,
            )
        else:
            return cls.from_chatbot(llm_settings.current_chatbot)

    @classmethod
    def from_default(cls) -> "LLMRequest":
        """
        Create an LLMRequest instance with default settings.

        Returns:
            LLMRequest: A new instance with default settings from LLMGeneratorSettings.
        """
        return cls.from_llm_settings()


@dataclass
class OpenrouterMistralRequest(LLMRequest):
    """
    A specialized LLMRequest for OpenRouter Mistral model.

    This class extends LLMRequest with parameters specific to OpenRouter's
    Mistral implementation.

    Attributes:
        max_tokens: Maximum number of tokens to generate.
        temperature: Temperature for sampling.
        seed: Random seed for reproducible generation.
        top_p: Keep the top tokens with cumulative probability >= top_p.
        top_k: Keep only top-k tokens with highest probability.
        frequency_penalty: Penalty for token frequency (range: [-2, 2]).
        presence_penalty: Penalty for token presence (range: [-2, 2]).
        repetition_penalty: Penalty for repeating tokens (range: [-2, 2]).
        logit_bias: Token bias to apply during generation.
        top_logprobs: Number of most likely tokens to return at each step.
        min_p: Minimum probability for token consideration.
        top_a: Parameter for nucleus sampling.
    """

    max_tokens: int = 256
    temperature: float = 0.1
    seed: int = 42
    top_p: float = 0.9
    top_k: int = 50
    frequency_penalty: float = 0  # Range: [-2, 2]
    presence_penalty: float = 0  # Range: [-2, 2]
    repetition_penalty: float = 0  # Range: [-2, 2]
    logit_bias: float = 0
    top_logprobs: int = 0
    min_p: float = 0
    top_a: int = 0

    def to_dict(self) -> Dict:
        """
        Convert the request parameters to a dictionary suitable for OpenRouter API calls.

        Ensures all values meet minimum thresholds.

        Returns:
            Dict: Dictionary representation of the request parameters for OpenRouter.
        """
        min_val = 0.0001
        frequency_penalty = max(self.frequency_penalty, min_val)
        presence_penalty = max(self.presence_penalty, min_val)
        repetition_penalty = max(self.repetition_penalty, min_val)
        top_p = max(self.top_p, min_val)
        temperature = max(self.temperature, min_val)

        data = {
            "max_tokens": self.max_tokens,
            "temperature": temperature,
            "seed": self.seed,
            "top_p": top_p,
            "top_k": self.top_k,
            "frequency_penalty": frequency_penalty,
            "presence_penalty": presence_penalty,
            "repetition_penalty": repetition_penalty,
            "logit_bias": self.logit_bias,
            "top_logprobs": self.top_logprobs,
            "min_p": self.min_p,
            "top_a": self.top_a,
        }
        return data



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'prompt_container.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QGridLayout, QHBoxLayout, QPlainTextEdit,
    QPushButton, QSizePolicy, QSpacerItem, QWidget)

class Ui_prompt_container_widget(object):
    def setupUi(self, prompt_container_widget):
        if not prompt_container_widget.objectName():
            prompt_container_widget.setObjectName(u"prompt_container_widget")
        prompt_container_widget.resize(433, 316)
        self.gridLayout = QGridLayout(prompt_container_widget)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(0, 0, 0, 0)
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.prompt = QPlainTextEdit(prompt_container_widget)
        self.prompt.setObjectName(u"prompt")

        self.horizontalLayout.addWidget(self.prompt)

        self.secondary_prompt = QPlainTextEdit(prompt_container_widget)
        self.secondary_prompt.setObjectName(u"secondary_prompt")

        self.horizontalLayout.addWidget(self.secondary_prompt)


        self.gridLayout.addLayout(self.horizontalLayout, 0, 0, 1, 2)

        self.pushButton = QPushButton(prompt_container_widget)
        self.pushButton.setObjectName(u"pushButton")
        icon = QIcon(QIcon.fromTheme(u"user-trash"))
        self.pushButton.setIcon(icon)

        self.gridLayout.addWidget(self.pushButton, 1, 0, 1, 1)

        self.horizontalSpacer = QSpacerItem(396, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout.addItem(self.horizontalSpacer, 1, 1, 1, 1)


        self.retranslateUi(prompt_container_widget)
        self.pushButton.clicked.connect(prompt_container_widget.handle_delete_prompt_clicked)

        QMetaObject.connectSlotsByName(prompt_container_widget)
    # setupUi

    def retranslateUi(self, prompt_container_widget):
        prompt_container_widget.setWindowTitle(QCoreApplication.translate("prompt_container_widget", u"Form", None))
        self.prompt.setPlaceholderText(QCoreApplication.translate("prompt_container_widget", u"Enter prompt", None))
        self.secondary_prompt.setPlaceholderText(QCoreApplication.translate("prompt_container_widget", u"Enter optional second prompt", None))
        self.pushButton.setText("")
    # retranslateUi




from airunner.gui.windows.setup_wizard.llama_license.templates.llama_license_ui import Ui_llama_license
from airunner.gui.windows.setup_wizard.user_agreement.agreement_page import AgreementPage


class LlamaLicense(AgreementPage):
    class_name_ = Ui_llama_license
    setting_key = "llama_license"



import unittest
from unittest.mock import patch, MagicMock

from PySide6.QtCore import Qt
from PySide6.QtGui import QColor, QPainter, QPixmap, QCursor, QPen, QBrush
from PySide6.QtWidgets import QApplication

from airunner.gui.cursors.circle_brush import circle_cursor


class TestCircleBrush(unittest.TestCase):
    """Test cases for the circle_brush cursor creation functionality."""

    @classmethod
    def setUpClass(cls):
        """Set up QApplication instance for all tests."""
        if not QApplication.instance():
            cls.qapp = QApplication([])
        else:
            cls.qapp = QApplication.instance()

    @classmethod
    def tearDownClass(cls):
        """Clean up QApplication instance after tests."""
        cls.qapp.quit()

    def test_circle_cursor_creates_qcursor(self):
        """Test that circle_cursor returns a QCursor object."""
        cursor = circle_cursor("black", "white")
        self.assertIsInstance(cursor, QCursor)

    def test_circle_cursor_with_default_size(self):
        """Test circle_cursor with default size."""
        cursor = circle_cursor("black", "white")
        pixmap = cursor.pixmap()
        self.assertEqual(pixmap.width(), 32)
        self.assertEqual(pixmap.height(), 32)

    def test_circle_cursor_with_custom_size(self):
        """Test circle_cursor with custom size."""
        cursor = circle_cursor("black", "white", pixmap_size=64)
        pixmap = cursor.pixmap()
        self.assertEqual(pixmap.width(), 64)
        self.assertEqual(pixmap.height(), 64)

    @patch('airunner.gui.cursors.circle_brush.QPainter')
    @patch('airunner.gui.cursors.circle_brush.QPixmap')
    @patch('airunner.gui.cursors.circle_brush.QCursor')
    def test_circle_cursor_drawing_operations(self, mock_qcursor, mock_qpixmap, mock_qpainter):
        """Test that drawing operations are performed correctly."""
        # Set up mocks
        mock_pixmap_instance = MagicMock()
        mock_qpixmap.return_value = mock_pixmap_instance

        mock_painter_instance = MagicMock()
        mock_qpainter.return_value = mock_painter_instance

        # Explicitly set the return value for RenderHint.Antialiasing
        mock_qpainter.RenderHint.Antialiasing = QPainter.RenderHint.Antialiasing

        mock_cursor_instance = MagicMock()
        mock_qcursor.return_value = mock_cursor_instance

        # Call the function
        result = circle_cursor("red", "blue", pixmap_size=48)

        # Verify pixmap creation
        mock_qpixmap.assert_called_once_with(48, 48)
        mock_pixmap_instance.fill.assert_called_once_with(Qt.GlobalColor.transparent)

        # Verify painter setup
        mock_qpainter.assert_called_once_with(mock_pixmap_instance)
        mock_painter_instance.setRenderHint.assert_called_once_with(QPainter.RenderHint.Antialiasing)

        # Verify drawing operations
        self.assertEqual(mock_painter_instance.setPen.call_count, 2)
        self.assertEqual(mock_painter_instance.setBrush.call_count, 2)
        self.assertEqual(mock_painter_instance.drawEllipse.call_count, 2)

        # Verify painter cleanup
        mock_painter_instance.end.assert_called_once()

        # Verify cursor creation and return
        mock_qcursor.assert_called_once_with(mock_pixmap_instance, 24, 24)
        self.assertEqual(result, mock_cursor_instance)

    def test_circle_cursor_with_different_colors(self):
        """Test circle_cursor with different outline and fill colors."""
        # Test with various color combinations
        color_pairs = [
            ("red", "blue"),
            ("green", "yellow"),
            ("#FF5733", "#C70039"),
            ("rgb(255,0,0)", "rgb(0,255,0)")
        ]
        
        for outline, fill in color_pairs:
            cursor = circle_cursor(outline, fill)
            self.assertIsInstance(cursor, QCursor)
            # We can't directly test the colors in the cursor, but at least we can verify it creates successfully

    def test_circle_cursor_hotspot(self):
        """Test that cursor hotspot is set correctly."""
        # Test hotspot for different sizes
        sizes = [16, 32, 48, 64]
        
        for size in sizes:
            cursor = circle_cursor("black", "white", pixmap_size=size)
            # The hotspot should be at the center of the pixmap
            expected_hotspot = size // 2
            
            # Get the actual hotspot - QCursor doesn't expose the hotspot directly
            # in a convenient way for testing, so we'll just infer from the size
            # This is more of a functional test ensuring the cursor is created correctly
            self.assertIsInstance(cursor, QCursor)
            pixmap = cursor.pixmap()
            self.assertEqual(pixmap.width(), size)
            self.assertEqual(pixmap.height(), size)

    def test_circle_cursor_integer_size(self):
        """Test that non-integer size is converted to integer."""
        # Test with a float size
        cursor = circle_cursor("black", "white", pixmap_size=32.5)
        pixmap = cursor.pixmap()
        self.assertEqual(pixmap.width(), 32)
        self.assertEqual(pixmap.height(), 32)
        
        # Test with a string size (should be converted to int if possible)
        with self.assertRaises(ValueError):
            circle_cursor("black", "white", pixmap_size="not-a-number")


if __name__ == "__main__":
    unittest.main()


from typing import Dict

from airunner.gui.widgets.nodegraph.nodes.art.base_art_node import (
    BaseArtNode,
)
from airunner.handlers.stablediffusion.image_request import ImageRequest
from airunner.enums import SignalCode
from airunner.vendor.nodegraphqt.constants import NodePropWidgetEnum


class GenerateImageNode(BaseArtNode):
    """
    Node for generating images using a specified Stable Diffusion model
    """

    NODE_NAME = "Generate Image"
    _input_ports = [
        {"name": "controlnet", "display_name": True},
        {"name": "image_to_image", "display_name": True},
        {"name": "image_request", "display_name": True},
    ]
    _output_ports = [
        {"name": "image_response", "display_name": True},
        {"name": "image", "display_name": True},
    ]
    _propertes = [
        {
            "name": "unload_after_generation",
            "value": True,
            "widget_type": NodePropWidgetEnum.QCHECK_BOX.value,
            "tab": "basic",
        }
    ]

    def __init__(self):
        self.signal_handlers = {
            SignalCode.ENGINE_RESPONSE_WORKER_RESPONSE_SIGNAL: self._on_image_generated
        }
        self._pending_request = False
        super().__init__()

    def _generate_image(self, image_request: ImageRequest):
        # Store the current node ID with the request to identify this node
        image_request.node_id = self.id
        self.api.art.send_request(image_request)

    def _on_image_generated(self, data: Dict):
        image_response = data.get("message", None)
        if image_response is None:
            self.api.nodegraph.node_executed(
                node_id=self.id,
                result=self.EXEC_OUT_PORT_NAME,
                data={"image_response": None, "image": None},
            )
            self._pending_request = False
            return

        if image_response.node_id is None or image_response.node_id != self.id:
            return

        self._pending_request = False

        output_data = {
            "image_response": image_response,
            "image": (
                image_response.images[0]
                if image_response and image_response.images
                else None
            ),
        }

        self.api.nodegraph.node_executed(
            node_id=self.id,
            result=self.EXEC_OUT_PORT_NAME,
            data=output_data,
        )

        if self.get_property("unload_after_generation"):
            self.api.art.unload()

    def execute(self, input_data: Dict):
        """
        Execute the node to generate an image.

        If this is the first execution, start image generation and return None
        to indicate pending execution.
        """
        # Check if we're already waiting for generation
        if self._pending_request:
            self.logger.warning(
                f"Node {self.id} is already waiting for image generation"
            )
            return None

        # Get image request from input
        image_request = input_data.get("image_request", None)
        if image_request is None:
            self.logger.error("No image request provided to GenerateImageNode")
            return {"image_response": None, "image": None}

        # Mark as pending and initiate image generation
        self.logger.info(f"Starting image generation for node {self.id}")
        image_request.node_id = self.id
        self._generate_image(image_request)
        self._pending_request = True

        # Return None to indicate the node execution is pending
        # The NodeGraphWorker will pause execution until NODE_EXECUTION_COMPLETED_SIGNAL
        return None

    def on_stop(self):
        super().on_stop()
        self.api.art.interrupt_generate()



import os
from sqlalchemy import Column, Integer, String

from airunner.data.models.base import BaseModel
from airunner.settings import AIRUNNER_BASE_PATH


class PathSettings(BaseModel):
    __tablename__ = 'path_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    base_path = Column(String, default=AIRUNNER_BASE_PATH)
    documents_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/other", "documents")))
    ebook_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/other", "ebooks")))
    image_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "art/other", "images")))
    llama_index_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/rag", "db")))
    webpages_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/other", "webpages")))
    stt_model_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/models/stt")))
    tts_model_path = Column(String, default=os.path.expanduser(os.path.join(AIRUNNER_BASE_PATH, "text/models/tts")))

    def tts_processor_path(self) -> str:
        return os.path.join(self.tts_model_path, "processor")



from PySide6.QtCore import Qt

def enable_wayland_window_decorations(window):
    """
    Enable window decorations for a Qt window when running in Wayland.
    This helps fix issues with windows having no borders/decorations in Docker/Wayland environments.
    
    :param window: A QWindow or QWidget instance
    """
    # Make sure window has correct flags for Wayland decorations
    window.setWindowFlags(
        Qt.WindowType.Window | 
        Qt.WindowType.WindowSystemMenuHint | 
        Qt.WindowType.WindowTitleHint | 
        Qt.WindowType.WindowMinMaxButtonsHint | 
        Qt.WindowType.WindowCloseButtonHint
    )


import queue
import time

import numpy as np
from PySide6.QtCore import QThread

from airunner.enums import SignalCode, ModelStatus
from airunner.settings import AIRUNNER_SLEEP_TIME_IN_MS
from airunner.workers.worker import Worker


class AudioCaptureWorker(Worker):
    """
    This class is responsible for capturing audio from the microphone.
    It will capture audio when it detects voice activity and then send the audio to the audio_processor_worker.
    """

    def __init__(self):
        self.signal_handlers = {
            SignalCode.AUDIO_CAPTURE_WORKER_RESPONSE_SIGNAL: self.on_audio_capture_worker_response_signal,
            SignalCode.STT_START_CAPTURE_SIGNAL: self.on_stt_start_capture_signal,
            SignalCode.STT_STOP_CAPTURE_SIGNAL: self.on_stt_stop_capture_signal,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.RECORDING_DEVICE_CHANGED: self.on_recording_device_changed_signal,
        }
        super().__init__()
        self.listening: bool = False
        self.voice_input_start_time: time.time = None
        self.chunk_duration = (
            self.stt_settings.chunk_duration
        )  # duration of chunks in milliseconds
        self.fs = self.stt_settings.fs
        self.running = False
        self._use_playback_stream: bool = False
        self._audio_process_queue = queue.Queue()

    @property
    def recording_device(self):
        """
        Get the recording device name from the sound settings.
        """
        recording_device = self.sound_settings.recording_device
        return recording_device if recording_device != "" else "pulse"

    def on_audio_capture_worker_response_signal(self, message: dict):
        item: np.ndarray = message["item"]
        self.logger.debug("Heard signal")
        self.add_to_queue(item)

    def on_stt_start_capture_signal(self):
        if not self.listening:
            self._start_listening()

    def on_stt_stop_capture_signal(self):
        if self.listening:
            self._stop_listening()

    def on_model_status_changed_signal(self, message: dict):
        model = message["model"]
        status = message["status"]
        if model == "stt" and status is ModelStatus.LOADED:
            self._start_listening()
        elif model == "stt" and status in (
            ModelStatus.UNLOADED,
            ModelStatus.FAILED,
        ):
            self._stop_listening()

    def run_thread(self):
        self.logger.debug("Starting audio capture worker")
        self.running = True
        chunk_duration = self.stt_settings.chunk_duration
        volume_input_threshold = self.stt_settings.volume_input_threshold
        silence_buffer_seconds = self.stt_settings.silence_buffer_seconds
        voice_input_start_time = None
        recording = []
        is_receiving_input = False

        while (
            self.listening
            and self.running
            and self.api.sounddevice_manager.in_stream
        ):
            try:
                # Use the actual sample rate from the stream if available
                actual_fs = (
                    self.api.sounddevice_manager.in_stream.samplerate
                    if hasattr(
                        self.api.sounddevice_manager.in_stream,
                        "samplerate",
                    )
                    else self.stt_settings.fs
                )
                frames = int(chunk_duration * actual_fs)
                chunk_data = self.api.sounddevice_manager.read_from_input(
                    frames
                )

                if chunk_data is None or chunk_data[0] is None:
                    QThread.msleep(AIRUNNER_SLEEP_TIME_IN_MS)
                    continue

                chunk, overflowed = chunk_data
                if chunk.ndim > 1:
                    chunk = np.mean(chunk, axis=1)
            except Exception as e:
                self.logger.error(f"Error reading from input stream: {e}")
                QThread.msleep(AIRUNNER_SLEEP_TIME_IN_MS)
                continue

            if (
                self._use_playback_stream
                and self.api.sounddevice_manager.out_stream
            ):
                try:
                    self.api.sounddevice_manager.write_to_output(chunk)
                except Exception as e:
                    self.logger.error(f"Playback error: {e}")

            if (
                np.max(np.abs(chunk)) > volume_input_threshold
            ):  # check if chunk is not silence
                self.logger.debug("Heard voice")

                is_receiving_input = True
                self.emit_signal(SignalCode.INTERRUPT_PROCESS_SIGNAL)
                voice_input_start_time = time.time()
            elif voice_input_start_time is not None:
                # make voice_end_time silence_buffer_seconds after voice_input_start_time
                end_time = voice_input_start_time + silence_buffer_seconds
                if time.time() >= end_time:
                    if len(recording) > 0:
                        self.logger.debug(
                            "Sending audio to audio_processor_worker"
                        )
                        self.emit_signal(
                            SignalCode.UNBLOCK_TTS_GENERATOR_SIGNAL,
                            {
                                "callback": lambda _recording=recording: self.emit_signal(
                                    SignalCode.AUDIO_CAPTURE_WORKER_RESPONSE_SIGNAL,
                                    {"item": b"".join(_recording)},
                                )
                            },
                        )
                        recording = []
                        is_receiving_input = False
            if is_receiving_input:
                chunk_bytes = np.int16(
                    chunk * 32767
                ).tobytes()  # convert to bytes
                recording.append(chunk_bytes)

        while not self.listening and self.running:
            QThread.msleep(AIRUNNER_SLEEP_TIME_IN_MS)

    def _start_listening(self):
        self.logger.debug("Start listening")
        self._initialize_stream()
        self.listening = True

    def _stop_listening(self):
        self.logger.debug("Stop listening")
        self.listening = False
        # Do not set self.running = False here - it terminates the worker
        # Allow proper cleanup to happen in the main loop

    def on_recording_device_changed_signal(self):
        self.logger.debug(f"Recording device changed")
        self._initialize_stream()

    def _initialize_stream(self):
        self.logger.debug("Initializing audio capture stream")
        samplerate = 16000  # self.stt_settings.fs
        channels = self.stt_settings.channels

        # Close any existing streams first to avoid conflicts
        if self.api.sounddevice_manager.in_stream:
            self.logger.debug(
                "Closing existing input stream before initializing a new one"
            )
            self.api.sounddevice_manager._stop_input_stream()

        # Log available input devices to help with debugging
        try:
            import sounddevice as sd

            devices = sd.query_devices()
            input_devices = [
                d for d in devices if d.get("max_input_channels", 0) > 0
            ]
            self.logger.debug(
                f"Available input devices: {[d['name'] for d in input_devices]}"
            )
            self.logger.debug(
                f"Selected recording device: {self.recording_device}"
            )
        except Exception as e:
            self.logger.error(f"Error querying audio devices: {e}")

        # Initialize input stream with better error handling
        success = self.api.sounddevice_manager.initialize_input_stream(
            samplerate=samplerate,
            channels=channels,
            device_name=self.recording_device,
        )

        if success:
            self.logger.info(
                f"Successfully initialized input stream with device: {self.recording_device}"
            )
        else:
            self.logger.error(
                f"Failed to initialize input stream with device: {self.recording_device}"
            )
            # Try with default device as fallback
            self.logger.warning("Attempting to initialize with default device")
            success = self.api.sounddevice_manager.initialize_input_stream(
                samplerate=samplerate,
                channels=channels,
                device_name="",  # Empty string should use system default
            )
            if success:
                self.logger.info(
                    "Successfully initialized input stream with default device"
                )
            else:
                self.logger.error(
                    "Failed to initialize input stream with default device"
                )

        # Initialize playback stream if needed
        if self._use_playback_stream:
            device_name = self.sound_settings.playback_device or "pulse"
            self.logger.debug(
                f"Initializing monitoring playback stream with device: {device_name}"
            )
            playback_success = (
                self.api.sounddevice_manager.initialize_output_stream(
                    samplerate=samplerate,
                    channels=channels,
                    device_name=device_name,
                )
            )
            if not playback_success:
                self.logger.error(
                    f"Failed to initialize playback stream with device: {device_name}"
                )

        # Verify stream status
        if self.api.sounddevice_manager.in_stream:
            self.logger.debug("Input stream is now active and ready")
        else:
            self.logger.error("Input stream failed to initialize properly")

        return self.api.sounddevice_manager.in_stream is not None



from PySide6.QtWidgets import QWizard
from PySide6.QtCore import QTimer
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.gui.windows.setup_wizard.installation_settings.install_failed_page import (
    InstallFailedPage,
)
from airunner.gui.windows.setup_wizard.installation_settings.choose_models_page import (
    ChooseModelsPage,
)
from airunner.gui.windows.setup_wizard.installation_settings.install_page import (
    InstallPage,
)
from airunner.gui.windows.setup_wizard.installation_settings.install_success_page import (
    InstallSuccessPage,
)
from airunner.gui.windows.setup_wizard.path_settings.path_settings import (
    PathSettings,
)


class DownloadWizardWindow(
    MediatorMixin,
    SettingsMixin,
    QWizard,
):
    """
    The download wizard window class for AI Runner.
    This class is used to download models and other resources required for AI Runner.
    """

    def __init__(self):
        """
        Initialize the download wizard window.
        :param setup_settings: The setup settings dictionary.
        """
        super().__init__()
        self.install_page = None
        self.setWindowTitle("AI Runner Download Wizard")
        self.setWizardStyle(QWizard.WizardStyle.ModernStyle)
        self.setOption(QWizard.WizardOption.IndependentPages, True)

        self.button(QWizard.WizardButton.FinishButton).clicked.connect(
            self.save_settings
        )

        self.button(QWizard.WizardButton.NextButton).clicked.connect(
            self.next_button_clicked
        )

        self.init_pages()

        # Connect signals for Next and Back buttons to control their state
        self.currentIdChanged.connect(self.on_page_changed)

        # We'll disable the next button immediately if we're on the installation page
        self.disableNextIfInstallPage()

    def save_settings(self):
        """
        Override this function to save settings based on specific page in question.
        Do not call this function directly.
        :return:
        """
        self.update_application_settings("run_setup_wizard", False)
        self.update_application_settings("download_wizard_completed", True)

    def next_button_clicked(self):
        current_page = self.currentPage()
        if hasattr(current_page, "next"):
            current_page.next()
        if hasattr(current_page, "start"):
            current_page.start()

    def init_pages(self):
        failed = True
        self.setPage(0, PathSettings(self))
        choose_models_page = ChooseModelsPage(self)
        self.setPage(1, choose_models_page)
        try:
            self.install_page = InstallPage(
                self,
                stablediffusion_models=choose_models_page.models,
                models_enabled=choose_models_page.models_enabled,
            )
            self.setPage(2, self.install_page)
        except BaseException as e:
            print(f"Exception creating InstallPage: {e}")
        try:
            self.setPage(3, InstallSuccessPage(self))
        except BaseException as e:
            print(f"Exception creating InstallSuccessPage: {e}")
        failed = False
        if failed:
            print("Setting InstallFailedPage")
            self.setPage(1, InstallFailedPage(self))

    def on_page_changed(self, id):
        """Handle page changes to properly manage button states"""
        # If we're now on the install page, disable the next button
        if self.currentPage() == self.install_page:
            self.disableNextButton()

            # Start the installation process which will re-enable the button when done
            if self.install_page:
                QTimer.singleShot(100, self.install_page.start)

    def disableNextIfInstallPage(self):
        """Check if current page is install page and disable Next button if so"""
        if (
            hasattr(self, "install_page")
            and self.currentPage() == self.install_page
        ):
            self.disableNextButton()

    def disableNextButton(self):
        """Disable the Next button"""
        if self.button(QWizard.WizardButton.NextButton):
            self.button(QWizard.WizardButton.NextButton).setEnabled(False)

    def show_final_page(self):
        """
        Show the final page.
        """
        # Find the ID of the InstallSuccessPage
        for page_id in self.pageIds():
            page = self.page(page_id)
            if isinstance(page, InstallSuccessPage):
                self.setCurrentId(page_id)
                # Don't call next() on previous_page as it doesn't exist
                # Instead, just update the wizard's UI state
                self.button(QWizard.WizardButton.BackButton).setEnabled(
                    False
                )  # Disable back button
                self.button(QWizard.WizardButton.NextButton).setEnabled(
                    False
                )  # Disable next button
                self.button(QWizard.WizardButton.FinishButton).setEnabled(
                    True
                )  # Enable finish button
                return



from typing import Any, List, Optional

from llama_index.core.base.llms.types import ChatMessage, MessageRole
from llama_index.core.memory import ChatMemoryBuffer as ChatMemoryBufferBroken


class ChatMemoryBuffer(ChatMemoryBufferBroken):
    def get(
        self, 
        input: Optional[str] = None, 
        initial_token_count: int = 0, 
        **kwargs: Any
    ) -> List[ChatMessage]:
        """Get chat history."""
        chat_history = self.get_all()

        if initial_token_count > self.token_limit:
            raise ValueError("Initial token count exceeds token limit")

        message_count = len(chat_history) if chat_history else 0
        if message_count > 0:
            cur_messages = chat_history[-message_count:]
            token_count = self._token_count_for_messages(cur_messages) + initial_token_count

            while token_count > self.token_limit and message_count > 1:
                message_count -= 1
                while chat_history[-message_count].role in (
                    MessageRole.TOOL,
                    MessageRole.ASSISTANT,
                ):
                    # we cannot have an assistant message at the start of the chat history
                    # if after removal of the first, we have an assistant message,
                    # we need to remove the assistant message too
                    #
                    # all tool messages should be preceded by an assistant message
                    # if we remove a tool message, we need to remove the assistant message too
                    message_count -= 1

                cur_messages = chat_history[-message_count:]
                token_count = (
                    self._token_count_for_messages(cur_messages) + initial_token_count
                )
            
            # catch one message longer than token limit
            if token_count > self.token_limit or message_count <= 0:
                return []

            return chat_history[-message_count:]
        return []



from sqlalchemy import Column, Integer, String

from airunner.data.models.base import BaseModel
from airunner.settings import AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR, AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR


class BrushSettings(BaseModel):
    __tablename__ = 'brush_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    size = Column(Integer, default=75)
    primary_color = Column(String, default=AIRUNNER_DEFAULT_BRUSH_PRIMARY_COLOR)
    secondary_color = Column(String, default=AIRUNNER_DEFAULT_BRUSH_SECONDARY_COLOR)
    strength_slider = Column(Integer, default=950)
    strength = Column(Integer, default=950)
    conditioning_scale = Column(Integer, default=550)
    guidance_scale = Column(Integer, default=75)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'choose_models.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QGridLayout, QGroupBox,
    QHBoxLayout, QLabel, QScrollArea, QSizePolicy,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_install_success_page(object):
    def setupUi(self, install_success_page):
        if not install_success_page.objectName():
            install_success_page.setObjectName(u"install_success_page")
        install_success_page.resize(615, 853)
        self.gridLayout = QGridLayout(install_success_page)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.page_title = QLabel(install_success_page)
        self.page_title.setObjectName(u"page_title")
        font = QFont()
        font.setPointSize(11)
        font.setBold(True)
        self.page_title.setFont(font)

        self.horizontalLayout.addWidget(self.page_title)

        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.total_size_label = QLabel(install_success_page)
        self.total_size_label.setObjectName(u"total_size_label")

        self.horizontalLayout.addWidget(self.total_size_label)


        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout.addItem(self.verticalSpacer, 9, 0, 1, 1)

        self.groupBox = QGroupBox(install_success_page)
        self.groupBox.setObjectName(u"groupBox")
        font1 = QFont()
        font1.setBold(True)
        self.groupBox.setFont(font1)
        self.groupBox.setCheckable(True)
        self.gridLayout_2 = QGridLayout(self.groupBox)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.stable_diffusion_scrollarea = QScrollArea(self.groupBox)
        self.stable_diffusion_scrollarea.setObjectName(u"stable_diffusion_scrollarea")
        self.stable_diffusion_scrollarea.setWidgetResizable(True)
        self.stable_diffusion_layout = QWidget()
        self.stable_diffusion_layout.setObjectName(u"stable_diffusion_layout")
        self.stable_diffusion_layout.setGeometry(QRect(0, 0, 571, 279))
        self.verticalLayout = QVBoxLayout(self.stable_diffusion_layout)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.stable_diffusion_scrollarea.setWidget(self.stable_diffusion_layout)

        self.gridLayout_2.addWidget(self.stable_diffusion_scrollarea, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox, 2, 0, 1, 1)

        self.checkBox = QCheckBox(install_success_page)
        self.checkBox.setObjectName(u"checkBox")
        self.checkBox.setChecked(True)

        self.gridLayout.addWidget(self.checkBox, 4, 0, 1, 1)

        self.speecht5_checkbox = QCheckBox(install_success_page)
        self.speecht5_checkbox.setObjectName(u"speecht5_checkbox")
        self.speecht5_checkbox.setChecked(True)

        self.gridLayout.addWidget(self.speecht5_checkbox, 5, 0, 1, 1)

        self.whisper_checkbox = QCheckBox(install_success_page)
        self.whisper_checkbox.setObjectName(u"whisper_checkbox")
        self.whisper_checkbox.setChecked(True)

        self.gridLayout.addWidget(self.whisper_checkbox, 7, 0, 1, 1)

        self.ministral_checkbox = QCheckBox(install_success_page)
        self.ministral_checkbox.setObjectName(u"ministral_checkbox")
        self.ministral_checkbox.setChecked(True)

        self.gridLayout.addWidget(self.ministral_checkbox, 3, 0, 1, 1)

        self.openvoice_model = QCheckBox(install_success_page)
        self.openvoice_model.setObjectName(u"openvoice_model")
        self.openvoice_model.setChecked(True)

        self.gridLayout.addWidget(self.openvoice_model, 6, 0, 1, 1)


        self.retranslateUi(install_success_page)
        self.ministral_checkbox.toggled.connect(install_success_page.ministral_toggled)
        self.speecht5_checkbox.toggled.connect(install_success_page.speecht5_toggled)
        self.whisper_checkbox.toggled.connect(install_success_page.whisper_toggled)
        self.groupBox.toggled.connect(install_success_page.stable_diffusion_toggled)
        self.checkBox.toggled.connect(install_success_page.embedding_model_toggled)
        self.openvoice_model.toggled.connect(install_success_page.openvoice_toggled)

        QMetaObject.connectSlotsByName(install_success_page)
    # setupUi

    def retranslateUi(self, install_success_page):
        install_success_page.setWindowTitle(QCoreApplication.translate("install_success_page", u"Form", None))
        self.page_title.setText(QCoreApplication.translate("install_success_page", u"Choose models to download", None))
        self.total_size_label.setText(QCoreApplication.translate("install_success_page", u"0MB", None))
        self.groupBox.setTitle(QCoreApplication.translate("install_success_page", u"Stable Diffusion Controlnet", None))
        self.checkBox.setText(QCoreApplication.translate("install_success_page", u"e5 Large: Embedding model (RAG search)", None))
        self.speecht5_checkbox.setText(QCoreApplication.translate("install_success_page", u"SpeechT5: Text-to-Speech", None))
        self.whisper_checkbox.setText(QCoreApplication.translate("install_success_page", u"Whisper: Speech-to-Text (voice conversations)", None))
        self.ministral_checkbox.setText(QCoreApplication.translate("install_success_page", u"Ministral 8B Instruct 4bit: Large Language Model (LLM)", None))
        self.openvoice_model.setText(QCoreApplication.translate("install_success_page", u"OpenVoice: Text-to-Speech", None))
    # retranslateUi




from airunner.gui.windows.about.templates.about_ui import Ui_about_window
from airunner.gui.windows.base_window import BaseWindow
# open the version file from the root of the project and get the VERSION variable string from it


class UpdateWindow(BaseWindow):
    template_class_ = Ui_about_window

    def initialize_window(self):
        current_text = self.template.current_version_label.text()
        latest_text = self.template.latest_version_label.text()
        self.template.current_version_label.setText(f"{current_text} {self.application_settings.app_version}")
        self.template.latest_version_label.setText(f"{latest_text} {self.application_settings.app_version}")



import unittest
from unittest.mock import patch, MagicMock

from PySide6.QtWidgets import QTabWidget, QApplication, QWidget

from airunner.data.models.tab import Tab


class TestTabModel(unittest.TestCase):
    """Test cases for the Tab model class."""

    @classmethod
    def setUpClass(cls):
        """Set up QApplication instance for the test suite."""
        if not QApplication.instance():
            cls.qapp = QApplication([])
        else:
            cls.qapp = QApplication.instance()

    @classmethod
    def tearDownClass(cls):
        """Clean up QApplication instance after tests."""
        cls.qapp.quit()

    def test_tab_model_attributes(self):
        """Test Tab model attributes and default values."""
        # Create a new Tab instance
        tab = Tab(
            section="",
            name="",
            active=False,
            displayed=True,
            index=0
        )
        
        # Verify attributes exist and have correct default values
        self.assertEqual(tab.section, "")
        self.assertEqual(tab.name, "")
        self.assertFalse(tab.active)
        self.assertTrue(tab.displayed)
        self.assertEqual(tab.index, 0)

    def test_tab_model_creation_with_values(self):
        """Test Tab model creation with specified values."""
        # Create a Tab with specific values
        tab = Tab(
            section="test_section",
            name="Test Tab",
            active=True,
            displayed=False,
            index=3
        )
        
        # Verify attributes have the specified values
        self.assertEqual(tab.section, "test_section")
        self.assertEqual(tab.name, "Test Tab")
        self.assertTrue(tab.active)
        self.assertFalse(tab.displayed)
        self.assertEqual(tab.index, 3)

    @patch('airunner.data.models.tab.session_scope')
    def test_update_tabs_sets_active_tab(self, mock_session_scope):
        """Test update_tabs method correctly updates active tab state."""
        # Create mock session and related objects
        mock_session = MagicMock()
        mock_query = MagicMock()
        mock_filter = MagicMock()
        mock_update = MagicMock()
        
        # Configure mocks for the session query chain
        mock_session_context = MagicMock()
        mock_session_context.__enter__.return_value = mock_session
        mock_session_scope.return_value = mock_session_context
        
        mock_session.query.return_value = mock_query
        mock_query.filter.return_value = mock_filter
        mock_filter.update.return_value = mock_update
        
        # Create a mock QTabWidget
        tab_widget = QTabWidget()
        tab_widget.addTab(QWidget(), "Tab 1")
        tab_widget.addTab(QWidget(), "Tab 2")
        tab_widget.addTab(QWidget(), "Tab 3")
        
        # Call the update_tabs method
        section = "test_section"
        index = 1  # Select "Tab 2"
        Tab.update_tabs(section, tab_widget, index)
        
        # Verify the first query sets all tabs in section to inactive
        first_call_args = mock_query.filter.call_args_list[0][0]
        self.assertEqual(len(first_call_args), 1)
        self.assertEqual(first_call_args[0].right.value, section)
        
        first_update_args = mock_filter.update.call_args_list[0][0]
        self.assertEqual(first_update_args[0], {"active": False})
        
        # Verify the second query sets the selected tab to active
        second_call_args = mock_query.filter.call_args_list[1][0]
        self.assertEqual(len(second_call_args), 2)
        self.assertEqual(second_call_args[0].right.value, section)
        self.assertEqual(second_call_args[1].right.value, "Tab 2")  # Tab text at index 1
        
        second_update_args = mock_filter.update.call_args_list[1][0]
        self.assertEqual(second_update_args[0], {"active": True})

    @patch('airunner.data.models.tab.session_scope')
    def test_update_tabs_with_empty_widget(self, mock_session_scope):
        """Test update_tabs with an empty QTabWidget."""
        # Create mock session
        mock_session = MagicMock()
        mock_session_context = MagicMock()
        mock_session_context.__enter__.return_value = mock_session
        mock_session_scope.return_value = mock_session_context
        
        # Create an empty QTabWidget
        tab_widget = QTabWidget()
        
        # Call the update_tabs method with an invalid index
        section = "test_section"
        index = 0  # Invalid index for empty widget
        
        # This should handle the error gracefully
        with self.assertRaises(IndexError):
            Tab.update_tabs(section, tab_widget, index)
        
        # Verify session.query was not called
        mock_session.query.assert_not_called()

    @patch('airunner.data.models.tab.session_scope')
    def test_update_tabs_with_invalid_index(self, mock_session_scope):
        """Test update_tabs with an invalid index."""
        # Create mock session
        mock_session = MagicMock()
        mock_session_context = MagicMock()
        mock_session_context.__enter__.return_value = mock_session
        mock_session_scope.return_value = mock_session_context
        
        # Create a QTabWidget with tabs
        tab_widget = QTabWidget()
        tab_widget.addTab(QWidget(), "Tab 1")
        tab_widget.addTab(QWidget(), "Tab 2")
        
        # Call the update_tabs method with an invalid index
        section = "test_section"
        index = 10  # Out of range
        
        # This should handle the error gracefully
        with self.assertRaises(IndexError):
            Tab.update_tabs(section, tab_widget, index)
        
        # Verify session.query was not called
        mock_session.query.assert_not_called()


if __name__ == "__main__":
    unittest.main()


from typing import Optional, Dict

from PySide6.QtCore import (
    QPointF,
    QPoint,
    Qt,
    QEvent,
    QSize,
    QTimer,
)
from PySide6.QtGui import QMouseEvent, QColor, QBrush
from PySide6.QtWidgets import (
    QGraphicsView,
    QGraphicsItemGroup,
    QGraphicsScene,
    QGraphicsTextItem,  # <-- add
    QInputDialog,  # <-- add
)

from airunner.enums import CanvasToolName, SignalCode, CanvasType
from airunner.gui.widgets.canvas.grid_graphics_item import GridGraphicsItem
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application.background_worker import BackgroundWorker
from airunner.utils.image import convert_image_to_binary
from airunner.gui.widgets.canvas.brush_scene import BrushScene
from airunner.gui.widgets.canvas.custom_scene import CustomScene
from airunner.gui.widgets.canvas.draggables.active_grid_area import (
    ActiveGridArea,
)
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.gui.widgets.canvas.zoom_handler import ZoomHandler
from airunner.utils.settings import get_qsettings


class CustomGraphicsView(
    MediatorMixin,
    SettingsMixin,
    QGraphicsView,
):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.setMouseTracking(True)
        self._initialized = False
        self._scene: Optional[CustomScene] = None
        self._canvas_color: str = "#000000"
        self.current_background_color: Optional[QColor] = None
        self.active_grid_area: Optional[ActiveGridArea] = None
        self.do_draw_layers: bool = True
        self.initialized: bool = False
        self.drawing: bool = False
        self.pixmaps: Dict = {}
        self.line_group: Optional[QGraphicsItemGroup] = None
        self._scene_is_active: bool = False
        self.last_pos: QPoint = self.zero_point
        self.zoom_handler: ZoomHandler = ZoomHandler()
        self.canvas_offset = QPointF(
            0, 0
        )  # Explicitly use QPointF, not QPoint
        self.settings = get_qsettings()
        self._middle_mouse_pressed: bool = False
        self.grid_item = None
        self._text_items = []  # Store references to QGraphicsTextItem
        self._editing_text_item = None

        # Add settings to handle negative coordinates properly
        self.setAlignment(
            Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft
        )  # Set once here
        self._last_viewport_size = self.viewport().size()  # Track last size
        self.setViewportUpdateMode(
            QGraphicsView.ViewportUpdateMode.BoundingRectViewportUpdate
        )

        # Use setOptimizationFlags directly instead of the enum that doesn't exist in your PySide6 version
        self.setOptimizationFlags(
            QGraphicsView.OptimizationFlag.DontAdjustForAntialiasing
            | QGraphicsView.OptimizationFlag.DontSavePainterState
        )

        # register signal handlers
        signal_handlers = {
            SignalCode.APPLICATION_TOOL_CHANGED_SIGNAL: self.on_tool_changed_signal,
            SignalCode.CANVAS_ZOOM_LEVEL_CHANGED: self.on_zoom_level_changed_signal,
            SignalCode.SET_CANVAS_COLOR_SIGNAL: self.set_canvas_color,
            SignalCode.UPDATE_SCENE_SIGNAL: self.update_scene,
            SignalCode.CANVAS_CLEAR_LINES_SIGNAL: self.clear_lines,
            SignalCode.SCENE_DO_DRAW_SIGNAL: self.on_canvas_do_draw_signal,
            SignalCode.APPLICATION_MAIN_WINDOW_LOADED_SIGNAL: self.on_main_window_loaded_signal,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal,
            SignalCode.MASK_GENERATOR_WORKER_RESPONSE_SIGNAL: self.on_mask_generator_worker_response_signal,
            SignalCode.RECENTER_GRID_SIGNAL: self.on_recenter_grid_signal,
            SignalCode.CANVAS_IMAGE_UPDATED_SIGNAL: self.on_canvas_image_updated_signal,
            SignalCode.CANVAS_UPDATE_IMAGE_POSITIONS: self.updateImagePositions,
        }
        for k, v in signal_handlers.items():
            self.register(k, v)

        self._pan_update_timer = QTimer()
        self._pan_update_timer.setSingleShot(True)
        self._pan_update_timer.timeout.connect(self._do_pan_update)
        self._pending_pan_event = False

        # Improved resize throttling
        self._resize_timer = QTimer()
        self._resize_timer.setSingleShot(True)
        self._resize_timer.timeout.connect(self._handle_deferred_resize)
        self._resize_data = None
        self._is_resizing = False

        self._cursor_cache = {}
        self._current_cursor = None

    @property
    def zero_point(self) -> QPointF:
        return QPointF(0, 0)  # Return QPointF instead of QPoint

    def load_canvas_offset(self):
        """Load the canvas offset from QSettings."""
        x = self.settings.value("canvas_offset_x", 0.0)  # Default to 0
        y = self.settings.value("canvas_offset_y", 0.0)  # Default to 0
        self.canvas_offset = QPointF(float(x), float(y))

        # Update positions after loading offset
        self.update_active_grid_area_position()
        self.updateImagePositions()
        self.do_draw()

    def save_canvas_offset(self):
        """Save the canvas offset to QSettings."""
        self.settings.setValue("canvas_offset_x", self.canvas_offset.x())
        self.settings.setValue("canvas_offset_y", self.canvas_offset.y())

    @property
    def scene(self) -> Optional[CustomScene]:
        scene = self._scene
        if not scene and self.canvas_type:
            if self.canvas_type == CanvasType.IMAGE.value:
                scene = CustomScene(canvas_type=self.canvas_type)
            elif self.canvas_type == CanvasType.BRUSH.value:
                scene = BrushScene(canvas_type=self.canvas_type)
            else:
                self.logger.error(f"Unknown canvas type: {self.canvas_type}")
                return

        if scene:
            scene.parent = self
            self._scene = scene
            self.setScene(scene)
            self.set_canvas_color(scene)
        return self._scene

    @scene.setter
    def scene(self, value: Optional[CustomScene]):
        self._scene = value

    @property
    def current_tool(self):
        val = getattr(self.application_settings, "current_tool", None)
        try:
            return CanvasToolName(val) if val is not None else None
        except Exception:
            return None

    @property
    def canvas_type(self) -> str:
        return self.property("canvas_type")

    @property
    def __do_show_active_grid_area(self):
        return self.canvas_type in (
            CanvasType.IMAGE.value,
            CanvasType.BRUSH.value,
        )

    @property
    def __can_draw_grid(self):
        return self.grid_settings.show_grid and self.canvas_type in (
            CanvasType.IMAGE.value,
            CanvasType.BRUSH.value,
        )

    def on_recenter_grid_signal(self):
        """Center the grid in the viewport with the active grid area's CENTER at the grid origin."""
        # 1. Calculate center of viewport
        viewport_size = self.viewport().size()
        viewport_center_x = viewport_size.width() / 2
        viewport_center_y = viewport_size.height() / 2

        # 2. To center the grid origin (0,0) in the viewport, set canvas offset to negative viewport center
        # This makes scene coordinate (0,0) appear at the center of the viewport
        self.canvas_offset = QPointF(-viewport_center_x, -viewport_center_y)
        self.save_canvas_offset()

        # 3. Calculate the position needed to center the active grid area on the origin (0,0)
        # We want the CENTER of the active grid area to be at (0,0), not its top-left corner
        grid_width = self.application_settings.working_width
        grid_height = self.application_settings.working_height

        # Position needs to be negative half-dimensions to center it
        pos_x = -grid_width / 2
        pos_y = -grid_height / 2

        # 4. Set active grid area to this centered position
        self.update_active_grid_settings("pos_x", int(pos_x))
        self.update_active_grid_settings("pos_y", int(pos_y))

        # 5. If there's an image in the scene, update its position to match the active grid area
        if self.scene and hasattr(self.scene, "item") and self.scene.item:
            # Store the same absolute position for the image as we did for the active grid area
            self.scene._original_item_positions[self.scene.item] = QPointF(
                pos_x, pos_y
            )

        # 6. Update all display positions based on new offset
        self.updateImagePositions()
        self.update_active_grid_area_position()
        self.update_drawing_pad_settings("x_pos", int(pos_x))
        self.update_drawing_pad_settings("y_pos", int(pos_y))
        self.do_draw(force_draw=True)

    def on_mask_generator_worker_response_signal(self, message: dict):
        mask = message["mask"]
        if mask is not None:
            mask = convert_image_to_binary(mask)
            self.update_drawing_pad_settings("mask", mask)

    def on_main_window_loaded_signal(self):
        self.initialized = True
        self.do_draw()

    def on_canvas_do_draw_signal(self, data: dict):
        self.do_draw(force_draw=data.get("force_draw", False))

    def on_application_settings_changed_signal(self, data: Dict):
        if data.get("setting_name") == "grid_settings":
            if (
                data.get("column_name") == "canvas_color"
                and self._canvas_color != data.get("value", self._canvas_color)
                and self.scene
            ):
                self._canvas_color = data.get("value")
                self.set_canvas_color(self.scene, self._canvas_color)

            if self.grid_settings.show_grid:
                self.do_draw()
            else:
                self.clear_lines()

    def do_draw(self, force_draw: bool = False, size: Optional[QSize] = None):
        if self.scene is None:
            return
        if (self.drawing or not self.initialized) and not force_draw:
            return
        self.drawing = True
        self.set_scene_rect()  # Set scene rect based on viewport

        # Remove old grid item if it exists
        if self.grid_item is not None:
            self.scene.removeItem(self.grid_item)
            self.grid_item = None

        # Add a single efficient grid item
        if self.grid_settings.show_grid:
            self.grid_item = GridGraphicsItem(self)
            self.scene.addItem(self.grid_item)

        self.show_active_grid_area()  # Ensure active grid is shown/positioned correctly
        self.update_scene()
        self.drawing = False

    def draw_grid(self, size: Optional[QSize] = None):
        if self.grid_item:
            self.grid_item.update()

    def clear_lines(self):
        if self.grid_item is not None:
            self.scene.removeItem(self.grid_item)
            self.grid_item = None

    def register_line_data(self, lines_data):
        for line_data in lines_data:
            try:
                line = self.scene.addLine(*line_data)
                self.line_group.addToGroup(line)
            except TypeError as e:
                self.logger.error(f"TypeError: {e}")
            except AttributeError as e:
                self.logger.error(f"AttributeError: {e}")

    def set_scene_rect(self):
        if not self.scene:
            return
        canvas_container_size = self.viewport().size()
        self.scene.setSceneRect(
            0, 0, canvas_container_size.width(), canvas_container_size.height()
        )

    def update_scene(self):
        if not self.scene:
            return
        self.scene.update()

    def remove_scene_item(self, item):
        if item is None:
            return
        if item.scene() == self.scene:
            self.scene.removeItem(item)

    def show_active_grid_area(self):
        if not self.__do_show_active_grid_area:
            # Ensure it's removed if disabled
            if self.active_grid_area:
                self.remove_scene_item(self.active_grid_area)
                self.active_grid_area = None
            return

        # Create if it doesn't exist
        if not self.active_grid_area:
            self.active_grid_area = ActiveGridArea()
            self.active_grid_area.setZValue(10)  # Ensure high visibility
            self.scene.addItem(self.active_grid_area)
            # Connect the signal emitted by the updated update_position
            self.active_grid_area.register(
                SignalCode.APPLICATION_ACTIVE_GRID_AREA_UPDATED,
                self.update_active_grid_area_position,  # Call view's update method
            )

        # Get the stored absolute position (defaults to 0,0 if not found)
        # Use active_grid_settings as the primary source, QSettings as fallback/persistence
        absolute_x = self.active_grid_settings.pos_x
        absolute_y = self.active_grid_settings.pos_y

        # If settings are somehow None (e.g., first run), default and save
        if absolute_x is None or absolute_y is None:
            # Default to centering in the initial view, considering the initial offset
            viewport_center_x = self.viewport().width() / 2
            viewport_center_y = self.viewport().height() / 2
            # Calculate absolute position needed to appear centered with current offset
            absolute_x = (
                viewport_center_x
                + self.canvas_offset.x()
                - (self.application_settings.working_width / 2)
            )
            absolute_y = (
                viewport_center_y
                + self.canvas_offset.y()
                - (self.application_settings.working_height / 2)
            )

            # Save this initial absolute position
            self.update_active_grid_settings("pos_x", int(round(absolute_x)))
            self.update_active_grid_settings("pos_y", int(round(absolute_y)))
            self.settings.sync()

        # Calculate and set the display position
        display_x = absolute_x - self.canvas_offset.x()
        display_y = absolute_y - self.canvas_offset.y()
        self.active_grid_area.setPos(display_x, display_y)

    def update_active_grid_area_position(
        self, _message=None
    ):  # Accept optional message from signal
        if self.active_grid_area:
            # Read the definitive absolute position from settings
            absolute_x = self.active_grid_settings.pos_x
            absolute_y = self.active_grid_settings.pos_y

            # Handle potential None values (shouldn't happen after show_active_grid_area)
            if absolute_x is None:
                absolute_x = 0
            if absolute_y is None:
                absolute_y = 0

            # Calculate the display position based on the current canvas offset
            display_x = absolute_x - self.canvas_offset.x()
            display_y = absolute_y - self.canvas_offset.y()

            # Set the item's display position
            self.active_grid_area.setPos(display_x, display_y)

    def on_zoom_level_changed_signal(self):
        transform = self.zoom_handler.on_zoom_level_changed()

        # Set the transform
        self.setTransform(transform)

        # Redraw lines
        self.do_draw()

    def resizeEvent(self, event):
        super().resizeEvent(event)

        # Store resize data
        self._resize_data = {
            "new_size": self.viewport().size(),
            "old_size": event.oldSize(),
        }

        # Start or reset the throttling timer - wait for resize to finish
        if not self._resize_timer.isActive():
            self._resize_timer.start(
                150
            )  # 150ms delay before processing resize
        else:
            # Reset the timer if already running
            self._resize_timer.stop()
            self._resize_timer.start(150)

    def _handle_deferred_resize(self):
        """Start a background thread to handle resize operations"""
        if not self._resize_data:
            return

        # Create a thread worker to handle the resize processing
        self._resize_worker = BackgroundWorker(
            task_function=self._process_resize,
            callback_data={"resize_data": self._resize_data},
        )

        # Connect the finished signal
        self._resize_worker.taskFinished.connect(self._on_resize_processed)

        # Clear resize data early to prevent double processing
        resize_data = self._resize_data
        self._resize_data = None

        # Start the background thread
        self._resize_worker.start()

    def _process_resize(self, worker=None):
        """Process resize in a background thread to prevent UI freezing

        Args:
            worker: The BackgroundWorker instance running this task (automatically provided)
        """
        # Get resize data from the worker's callback_data
        resize_data = (
            worker.callback_data.get("resize_data") if worker else None
        )

        if not resize_data:
            return None

        new_size = resize_data["new_size"]
        old_size = resize_data["old_size"]

        # Only proceed if we have valid sizes
        if (
            old_size.width() <= 0
            or old_size.height() <= 0
            or new_size.width() <= 0
            or new_size.height() <= 0
        ):
            return None

        # Calculate central reference point
        canvas_center_point = QPointF(
            new_size.width() / 2, new_size.height() / 2
        )

        # Calculate size differences
        delta_width = new_size.width() - old_size.width()
        delta_height = new_size.height() - old_size.height()

        # If only width changes and height is nearly the same, treat as splitter resize (no offset change)
        if abs(delta_height) < 5 and abs(delta_width) > 0:
            new_offset = self.canvas_offset  # Do not change offset
        else:
            # For window resize, keep center fixed
            new_offset = QPointF(
                self.canvas_offset.x() - delta_width / 2,
                self.canvas_offset.y() - delta_height / 2,
            )

        # Return the calculated values
        return {
            "canvas_offset": new_offset,
            "new_size": new_size,
            "canvas_center_point": canvas_center_point,
        }

    def _on_resize_processed(self, data):
        """Handle the completion of resize processing"""
        if "error" in data:
            self.logger.error(
                f"Error during resize processing: {data['error']}"
            )
            return

        if "result" not in data:
            return

        result = data["result"]
        if not result:
            return

        # Update canvas offset with the thread-calculated value
        self.canvas_offset = result["canvas_offset"]
        self._canvas_center_point = result["canvas_center_point"]
        self._last_viewport_size = result["new_size"]

        # Update positions on UI thread
        self.update_active_grid_area_position()
        self.updateImagePositions()

        # Redraw grid
        self.draw_grid()

    def wheelEvent(self, event):
        # Only allow zooming with Ctrl, otherwise ignore scrolling
        if event.modifiers() & Qt.KeyboardModifier.ControlModifier:
            super().wheelEvent(event)
            self.draw_grid()  # Only redraw grid on zoom
        else:
            event.ignore()  # Prevent QGraphicsView from scrolling

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self._middle_mouse_pressed = True
            self.last_pos = event.pos()
            event.accept()
            return
        # Only handle text tool logic if tool is TEXT
        if self.current_tool is CanvasToolName.TEXT:
            scene_pos = self.mapToScene(event.pos())
            item = self._find_text_item_at(scene_pos)
            if item:
                # Select and edit existing text item
                self._edit_text_item(item)
                return
            else:
                # Add new text item
                self._add_text_item(scene_pos)
                return
        # If not text tool, ensure all text items are not movable/editable
        self._set_text_items_interaction(False)
        super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        if event.button() == Qt.MouseButton.MiddleButton:
            self.save_canvas_offset()
            self._middle_mouse_pressed = False
            self.last_pos = None

            # After releasing middle mouse button, trigger a cursor update
            # Pass a fake enter event to the scene to refresh the cursor
            if self.scene:
                # Create a simple "dummy" event just to trigger cursor update
                class SimpleEvent:
                    def __init__(self):
                        pass

                    def type(self):
                        return QEvent.Type.Enter

                # Tell the scene to update the cursor based on current tool
                self.scene._handle_cursor(SimpleEvent(), True)

            event.accept()
            return
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event):
        if self._middle_mouse_pressed:
            delta = event.pos() - self.last_pos
            self.canvas_offset -= delta
            self.last_pos = event.pos()
            if not self._pan_update_timer.isActive():
                self._pan_update_timer.start(1)
            else:
                self._pending_pan_event = True
            event.accept()
            return
        super().mouseMoveEvent(event)

    def _do_pan_update(self):
        self.update_active_grid_area_position()
        self.updateImagePositions()
        self.draw_grid()
        if self._pending_pan_event:
            self._pending_pan_event = False
            self._pan_update_timer.start(1)

    def showEvent(self, event):
        super().showEvent(event)
        # Load offset first
        self.load_canvas_offset()

        # Set up the scene (grid, etc.)
        self.do_draw(True)
        self.toggle_drag_mode()
        self.set_canvas_color(self.scene)

        # Show the active grid area using loaded offset
        self.show_active_grid_area()
        self.scene.initialize_image()
        self.updateImagePositions()
        self._restore_text_items_from_db()  # Restore text items on load

        if not self._initialized:
            self._initialized = True

    def set_canvas_color(
        self,
        scene: Optional[CustomScene] = None,
        canvas_color: Optional[str] = None,
    ):
        scene = self.scene if not scene else scene
        canvas_color = canvas_color or self.grid_settings.canvas_color
        self.current_background_color = canvas_color
        color = QColor(self.current_background_color)
        brush = QBrush(color)
        scene.setBackgroundBrush(brush)

    def on_tool_changed_signal(self, message):
        self.toggle_drag_mode()
        # Update text item interaction flags based on tool
        is_text = self.current_tool is CanvasToolName.TEXT
        self._set_text_items_interaction(is_text)

    def toggle_drag_mode(self):
        self.setDragMode(QGraphicsView.DragMode.NoDrag)

    def handle_pan_canvas(self, event: QMouseEvent):
        if self._middle_mouse_pressed:
            delta = event.pos() - self.last_pos

            self.canvas_offset -= delta
            self.last_pos = event.pos()

            # Update display positions based on the NEW offset
            self.update_active_grid_area_position()
            self.updateImagePositions()
            self.draw_grid()

    def update_active_grid_area_position(self):
        if self.active_grid_area:
            pos = self.active_grid_settings.pos
            pos_x = pos[0] - self.canvas_offset.x()
            pos_y = pos[1] - self.canvas_offset.y()
            if self.active_grid_area:
                self.active_grid_area.setPos(pos_x, pos_y)

    def updateImagePositions(self):
        """Update positions of all images in the scene based on canvas offset."""
        if (
            not self.scene
            or not hasattr(self.scene, "item")
            or not self.scene.item
        ):
            return

        # Get the main item directly
        item = self.scene.item

        # Make sure the item is visible
        item.setVisible(True)

        # Store original position if needed
        if item not in self.scene._original_item_positions:
            self.scene._original_item_positions[item] = item.pos()

        # Get original position
        original_pos = self.scene._original_item_positions[item]

        # Calculate new position
        new_x = original_pos.x() - self.canvas_offset.x()
        new_y = original_pos.y() - self.canvas_offset.y()

        # Set new position
        item.setPos(new_x, new_y)

        # Ensure image has high Z value for visibility
        item.setZValue(5)

        # Ensure the item's bounding rect is properly updated
        item.prepareGeometryChange()

        # Force the scene to update the entire viewport
        self.scene.invalidate(
            item.boundingRect(), QGraphicsScene.SceneLayer.ItemLayer
        )

        # Force entire viewport update to handle negative coordinates
        self.viewport().update()

    def enterEvent(self, event: QEvent) -> None:
        """
        Handle the event when the mouse enters the CustomGraphicsView widget.
        Let the scene handle the cursor based on the current tool.
        """
        self.scene.enterEvent(event)
        super().enterEvent(event)
        # Remove the forced crosshair cursor to let the custom cursor logic work
        # self.setCursor(Qt.CursorShape.CrossCursor)  # This was forcing the plus sign cursor

    def leaveEvent(self, event: QEvent) -> None:
        """
        Handle the event when the mouse leaves the CustomGraphicsView widget.
        Resets the cursor to a normal pointer.
        """
        self.scene.leaveEvent(event)
        super().leaveEvent(event)

    def on_canvas_image_updated_signal(self, *args):
        """Handler for when images are updated or added to the canvas.
        Ensures that newly generated images respect the current pan position.
        """
        # Update image positions based on current canvas offset
        self.updateImagePositions()

        # Ensure the scene updates to show the new image correctly
        self.update_scene()

    def get_cached_cursor(self, tool, size):
        key = (tool, size)
        if key not in self._cursor_cache:
            from airunner.gui.cursors.circle_brush import circle_cursor

            if tool in (CanvasToolName.BRUSH, CanvasToolName.ERASER):
                # You may want to use different colors for eraser
                cursor = circle_cursor(
                    Qt.GlobalColor.white, Qt.GlobalColor.transparent, size
                )
                self._cursor_cache[key] = cursor
        return self._cursor_cache.get(key)

    def _update_cursor(self, event=None, current_tool=None, apply_cursor=True):
        # event: QEvent or similar, current_tool: CanvasToolName
        # apply_cursor: bool
        if current_tool is None:
            current_tool = self.current_tool
        cursor = None
        if apply_cursor:
            if (
                event
                and hasattr(event, "button")
                and event.button() == Qt.MouseButton.MiddleButton
            ):
                cursor = Qt.CursorShape.ClosedHandCursor
            elif current_tool in (CanvasToolName.BRUSH, CanvasToolName.ERASER):
                size = getattr(self, "brush_settings", None)
                size = size.size if size else 32
                cursor = self.get_cached_cursor(current_tool, size)
            elif current_tool is CanvasToolName.TEXT:
                cursor = Qt.CursorShape.IBeamCursor
            elif current_tool is CanvasToolName.ACTIVE_GRID_AREA:
                if (
                    event
                    and hasattr(event, "buttons")
                    and event.buttons() == Qt.MouseButton.LeftButton
                ):
                    cursor = Qt.CursorShape.ClosedHandCursor
                else:
                    cursor = Qt.CursorShape.OpenHandCursor
            elif current_tool is CanvasToolName.NONE:
                cursor = Qt.CursorShape.ArrowCursor
        else:
            cursor = Qt.CursorShape.ArrowCursor
        if self._current_cursor != cursor:
            self.setCursor(cursor)
            self._current_cursor = cursor

    def on_undo_button_clicked(self):
        self.api.art.canvas.undo()
        # Do not sync image to grid area; keep image at its last location

    def on_redo_button_clicked(self):
        self.api.art.canvas.redo()
        # Do not sync image to grid area; keep image at its last location

    def _find_text_item_at(self, pos):
        if not self.scene:
            return None
        items = self.scene.items(pos)
        for item in items:
            if isinstance(item, QGraphicsTextItem):
                return item
        return None

    def _add_text_item(self, pos):
        if not self.scene:
            return
        text, ok = QInputDialog.getText(self, "Add Text", "Enter text:")
        if ok and text:
            text_item = QGraphicsTextItem(text)
            text_item.setTextInteractionFlags(Qt.TextEditorInteraction)
            text_item.setFlag(QGraphicsTextItem.ItemIsMovable, True)
            text_item.setFlag(QGraphicsTextItem.ItemIsSelectable, True)
            text_item.setFlag(QGraphicsTextItem.ItemIsFocusable, True)
            text_item.setPos(pos)
            self.scene.addItem(text_item)
            self._text_items.append(text_item)
            text_item.focusOutEvent = self._make_text_focus_out_handler(
                text_item
            )
            text_item.keyPressEvent = self._make_text_key_press_handler(
                text_item
            )
            text_item.setFocus()
            self._editing_text_item = text_item
            text_item.setZValue(20)
            text_item.setDefaultTextColor(QColor("white"))
            text_item.setFont(self._get_default_text_font())
            text_item.setFlag(QGraphicsTextItem.ItemSendsGeometryChanges, True)
            text_item.itemChange = self._make_text_item_change_handler(
                text_item
            )
            self._save_text_items_to_db()

    def _edit_text_item(self, item):
        # Only allow editing if text tool is active
        if self.current_tool is CanvasToolName.TEXT:
            item.setTextInteractionFlags(Qt.TextEditorInteraction)
            item.setFlag(QGraphicsTextItem.ItemIsMovable, True)
            item.setFlag(QGraphicsTextItem.ItemIsSelectable, True)
            item.setFocus()
            self._editing_text_item = item

    def _make_text_focus_out_handler(self, item):
        def handler(event):
            item.setTextInteractionFlags(Qt.NoTextInteraction)
            self._editing_text_item = None
            self._save_text_items_to_db()
            QGraphicsTextItem.focusOutEvent(item, event)

        return handler

    def _make_text_key_press_handler(self, item):
        def handler(event):
            if event.key() == Qt.Key.Key_Delete:
                self.scene.removeItem(item)
                if item in self._text_items:
                    self._text_items.remove(item)
                self._save_text_items_to_db()
                self._editing_text_item = None
            else:
                QGraphicsTextItem.keyPressEvent(item, event)

        return handler

    def _make_text_item_change_handler(self, item):
        def handler(change, value):
            if change == QGraphicsTextItem.ItemPositionChange:
                self._save_text_items_to_db()
            return QGraphicsTextItem.itemChange(item, change, value)

        return handler

    def _get_default_text_font(self):
        from PySide6.QtGui import QFont

        font = QFont()
        font.setPointSize(18)
        font.setFamily("Arial")
        return font

    def _save_text_items_to_db(self):
        # Save all text items to the database (via application_settings or similar)
        text_items_data = []
        for item in self._text_items:
            text_items_data.append(
                {
                    "text": item.toPlainText(),
                    "x": item.pos().x(),
                    "y": item.pos().y(),
                    "color": item.defaultTextColor().name(),
                    "font": item.font().toString(),
                }
            )
        # Store in application_settings (or replace with actual DB call)
        self.update_drawing_pad_settings("text_items", text_items_data)

    def _restore_text_items_from_db(self):
        # Restore text items from the database (via application_settings or similar)
        self._clear_text_items()
        text_items_data = getattr(self.drawing_pad_settings, "text_items", [])
        from PySide6.QtGui import QFont

        for data in text_items_data:
            text = data.get("text", "")
            x = data.get("x", 0)
            y = data.get("y", 0)
            color = QColor(data.get("color", "white"))
            font = QFont()
            font.fromString(data.get("font", ""))
            text_item = QGraphicsTextItem(text)
            text_item.setPos(QPointF(x, y))
            text_item.setDefaultTextColor(color)
            text_item.setFont(font)
            text_item.setFlag(QGraphicsTextItem.ItemIsMovable, True)
            text_item.setFlag(QGraphicsTextItem.ItemIsSelectable, True)
            text_item.setFlag(QGraphicsTextItem.ItemIsFocusable, True)
            text_item.setFlag(QGraphicsTextItem.ItemSendsGeometryChanges, True)
            text_item.setZValue(20)
            text_item.focusOutEvent = self._make_text_focus_out_handler(
                text_item
            )
            text_item.keyPressEvent = self._make_text_key_press_handler(
                text_item
            )
            text_item.itemChange = self._make_text_item_change_handler(
                text_item
            )
            self.scene.addItem(text_item)
            self._text_items.append(text_item)

    def _clear_text_items(self):
        for item in self._text_items:
            self.scene.removeItem(item)
        self._text_items.clear()

    def _set_text_items_interaction(self, enable: bool):
        # Enable/disable moving/editing for all text items
        for item in self._text_items:
            if enable:
                item.setFlag(QGraphicsTextItem.ItemIsMovable, True)
                item.setFlag(QGraphicsTextItem.ItemIsSelectable, True)
            else:
                item.setFlag(QGraphicsTextItem.ItemIsMovable, False)
                item.setFlag(QGraphicsTextItem.ItemIsSelectable, False)
                item.setTextInteractionFlags(Qt.NoTextInteraction)



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'blend_option_widget.ui'
##
## Created by: Qt User Interface Compiler version 6.9.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDoubleSpinBox, QFormLayout,
    QFrame, QGridLayout, QHBoxLayout, QLabel,
    QPlainTextEdit, QPushButton, QSizePolicy, QSlider,
    QSpacerItem, QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(290, 940)
        self.verticalLayout_2 = QVBoxLayout(Form)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.gridLayout_2 = QGridLayout()
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.weight_slider = QSlider(Form)
        self.weight_slider.setObjectName(u"weight_slider")
        self.weight_slider.setMaximum(100)
        self.weight_slider.setSliderPosition(100)
        self.weight_slider.setOrientation(Qt.Horizontal)

        self.horizontalLayout.addWidget(self.weight_slider)

        self.line = QFrame(Form)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.HLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout.addWidget(self.line)

        self.weight_spinbox = QDoubleSpinBox(Form)
        self.weight_spinbox.setObjectName(u"weight_spinbox")
        self.weight_spinbox.setMaximum(1.000000000000000)
        self.weight_spinbox.setSingleStep(0.010000000000000)
        self.weight_spinbox.setValue(1.000000000000000)

        self.horizontalLayout.addWidget(self.weight_spinbox)


        self.gridLayout_2.addLayout(self.horizontalLayout, 3, 0, 1, 2)

        self.plainTextEdit = QPlainTextEdit(Form)
        self.plainTextEdit.setObjectName(u"plainTextEdit")

        self.gridLayout_2.addWidget(self.plainTextEdit, 1, 0, 1, 2)

        self.image_form = QFormLayout()
        self.image_form.setObjectName(u"image_form")
        self.image_form.setFormAlignment(Qt.AlignLeading|Qt.AlignLeft|Qt.AlignTop)
        self.image_container = QLabel(Form)
        self.image_container.setObjectName(u"image_container")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.image_container.sizePolicy().hasHeightForWidth())
        self.image_container.setSizePolicy(sizePolicy)
        self.image_container.setMinimumSize(QSize(128, 128))
        self.image_container.setMaximumSize(QSize(128, 128))
        self.image_container.setStyleSheet(u"border: 1px solid black")

        self.image_form.setWidget(0, QFormLayout.ItemRole.LabelRole, self.image_container)

        self.verticalLayout = QVBoxLayout()
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.layer_combobox = QComboBox(Form)
        self.layer_combobox.setObjectName(u"layer_combobox")

        self.verticalLayout.addWidget(self.layer_combobox)

        self.import_image_button = QPushButton(Form)
        self.import_image_button.setObjectName(u"import_image_button")

        self.verticalLayout.addWidget(self.import_image_button)

        self.image_spacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout.addItem(self.image_spacer)


        self.image_form.setLayout(0, QFormLayout.ItemRole.FieldRole, self.verticalLayout)


        self.gridLayout_2.addLayout(self.image_form, 2, 0, 1, 2)

        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_2.addItem(self.verticalSpacer, 5, 1, 1, 1)

        self.image_or_text_combobox = QComboBox(Form)
        self.image_or_text_combobox.addItem("")
        self.image_or_text_combobox.addItem("")
        self.image_or_text_combobox.setObjectName(u"image_or_text_combobox")

        self.gridLayout_2.addWidget(self.image_or_text_combobox, 0, 0, 1, 2)

        self.delete_button = QPushButton(Form)
        self.delete_button.setObjectName(u"delete_button")

        self.gridLayout_2.addWidget(self.delete_button, 4, 0, 1, 2)


        self.verticalLayout_2.addLayout(self.gridLayout_2)

        self.verticalSpacer_2 = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.verticalLayout_2.addItem(self.verticalSpacer_2)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.image_container.setText("")
        self.import_image_button.setText(QCoreApplication.translate("Form", u"Import new image", None))
        self.image_or_text_combobox.setItemText(0, QCoreApplication.translate("Form", u"Image", None))
        self.image_or_text_combobox.setItemText(1, QCoreApplication.translate("Form", u"Text", None))

        self.delete_button.setText(QCoreApplication.translate("Form", u"Delete", None))
    # retranslateUi




from PySide6.QtCore import Slot
from airunner.gui.windows.setup_wizard.base_wizard import BaseWizard
from airunner.gui.windows.setup_wizard.model_setup.stable_diffusion_setup.templates.stable_diffusion_model_setup_ui import \
    Ui_stable_diffusion_model_setup


class StableDiffusionSetupPage(BaseWizard):
    class_name_ = Ui_stable_diffusion_model_setup

    def __init__(self, *args):
        super(StableDiffusionSetupPage, self).__init__(*args)

        self



import os

os.environ["AIRUNNER_DB_NAME"] = "test.db"

import unittest
from unittest.mock import patch, MagicMock, PropertyMock
from PIL import Image

from PySide6.QtWidgets import QApplication

from airunner.handlers.stablediffusion.stable_diffusion_model_manager import (
    StableDiffusionModelManager,
)
from airunner.handlers.stablediffusion.image_request import ImageRequest
from airunner.enums import (
    ModelStatus,
    ModelType,
    HandlerState,
    SignalCode,
    EngineResponseCode,
    GeneratorSection,
)

import airunner.setup_database as setup_database

# Set test environment variables


class TestStableDiffusionModelManager(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # Initialize the QApplication for PySide6
        cls.app = QApplication.instance()
        if not cls.app:
            cls.app = QApplication([])

        # Setup the test database
        setup_database.setup_database()

    def setUp(self):
        # Mock dependencies
        self.mock_mediator = MagicMock()
        self.mock_path_settings = MagicMock()
        self.mock_path_settings.base_path = "/test/path"

        # Mock logger before creating the handler
        self.mock_logger = MagicMock()
        self.mock_model_status = {
            ModelType.SD: ModelStatus.UNLOADED,
            ModelType.SAFETY_CHECKER: ModelStatus.UNLOADED,
            ModelType.FEATURE_EXTRACTOR: ModelStatus.UNLOADED,
            ModelType.CONTROLNET: ModelStatus.UNLOADED,
            ModelType.SCHEDULER: ModelStatus.UNLOADED,
        }

        # Create a StableDiffusionModelManager instance with mocked dependencies
        with patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.torch.Generator"
        ) as mock_generator, patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.StableDiffusionSafetyChecker"
        ) as mock_safety_checker, patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.CLIPFeatureExtractor"
        ) as mock_feature_extractor, patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.ControlNetModel"
        ) as mock_controlnet_model, patch(
            "airunner.handlers.base_model_manager.BaseModelManager.__init__"
        ) as mock_base_init, patch.object(
            StableDiffusionModelManager,
            "path_settings_cached",
            self.mock_path_settings,
            create=True,
        ), patch.object(
            StableDiffusionModelManager,
            "logger",
            self.mock_logger,
            create=True,
        ):

            mock_base_init.return_value = None
            mock_generator.return_value = MagicMock()

            self.handler = StableDiffusionModelManager()

            # Setup necessary properties on the handler
            self.handler._model_status = self.mock_model_status
            self.handler.change_model_status = MagicMock()

    def tearDown(self):
        # Clean up after test
        self.handler = None

    def test_initialization(self):
        """Test the initialization of StableDiffusionModelManager"""
        self.assertEqual(self.handler._model_path, "/test/model/path")
        self.assertEqual(self.handler._model_version, "SD 1.5")
        self.assertEqual(self.handler._pipeline, "txt2img")
        self.assertEqual(
            self.handler._scheduler_name, "DPMSolverMultistepScheduler"
        )
        self.assertEqual(self.handler._use_compel, True)
        self.assertEqual(
            self.handler._current_state, HandlerState.UNINITIALIZED
        )
        self.assertEqual(self.handler.model_type, ModelType.SD)

    def test_properties(self):
        """Test the properties of StableDiffusionModelManager"""
        self.assertEqual(self.handler.model_path, "/test/model/path")

        # Test model_status property
        self.assertEqual(self.handler.model_status, self.mock_model_status)

        # Test scheduler_name property
        self.assertEqual(
            self.handler.scheduler_name, "DPMSolverMultistepScheduler"
        )

        # Test use_compel property
        self.assertEqual(self.handler.use_compel, True)

        # Test version property
        self.assertEqual(self.handler.version, "SD 1.5")

        # Test is_txt2img property
        with patch.object(
            StableDiffusionModelManager, "section", new_callable=PropertyMock
        ) as mock_section:
            mock_section.return_value = GeneratorSection.TXT2IMG
            self.assertTrue(self.handler.is_txt2img)

    def test_image_request_property(self):
        """Test the image_request property setter and getter"""
        mock_request = MagicMock(spec=ImageRequest)
        self.handler.image_request = mock_request
        self.assertEqual(self.handler.image_request, mock_request)

    def test_load_safety_checker(self):
        """Test the load_safety_checker method"""
        with patch.object(
            self.handler, "_load_safety_checker"
        ) as mock_load, patch.object(
            StableDiffusionModelManager,
            "safety_checker_is_loading",
            new_callable=PropertyMock,
        ) as mock_loading:
            # Test that it doesn't call _load_safety_checker when safety_checker_is_loading is True
            mock_loading.return_value = True
            self.handler.load_safety_checker()
            mock_load.assert_not_called()
            mock_load.reset_mock()

            # Test that it calls _load_safety_checker when safety_checker_is_loading is False
            mock_loading.return_value = False
            self.handler.load_safety_checker()
            mock_load.assert_called_once()

    def test_unload_safety_checker(self):
        """Test the unload_safety_checker method"""
        with patch.object(
            self.handler, "_unload_safety_checker"
        ) as mock_unload, patch.object(
            StableDiffusionModelManager,
            "safety_checker_is_loading",
            new_callable=PropertyMock,
        ) as mock_loading:
            # Test that it doesn't call _unload_safety_checker when safety_checker_is_loading is True
            mock_loading.return_value = True
            self.handler.unload_safety_checker()
            mock_unload.assert_not_called()
            mock_unload.reset_mock()

            # Test that it calls _unload_safety_checker when safety_checker_is_loading is False
            mock_loading.return_value = False
            self.handler.unload_safety_checker()
            mock_unload.assert_called_once()

    def test_load_controlnet(self):
        """Test the load_controlnet method"""
        with patch.object(
            self.handler, "_load_controlnet"
        ) as mock_load, patch.object(
            StableDiffusionModelManager,
            "controlnet_enabled",
            new_callable=PropertyMock,
        ) as mock_enabled, patch.object(
            StableDiffusionModelManager,
            "controlnet_is_loading",
            new_callable=PropertyMock,
        ) as mock_loading:
            # Don't load when controlnet is not enabled
            mock_enabled.return_value = False
            mock_loading.return_value = False
            self.handler.load_controlnet()
            mock_load.assert_not_called()
            mock_load.reset_mock()

            # Don't load when controlnet is already loading
            mock_enabled.return_value = True
            mock_loading.return_value = True
            self.handler.load_controlnet()
            mock_load.assert_not_called()
            mock_load.reset_mock()

            # Load when controlnet is enabled and not already loading
            mock_enabled.return_value = True
            mock_loading.return_value = False
            self.handler.load_controlnet()
            self.assertIsNone(self.handler._controlnet_model)
            self.assertIsNone(self.handler._controlnet_settings)
            mock_load.assert_called_once()

    def test_unload_controlnet(self):
        """Test the unload_controlnet method"""
        with patch.object(
            self.handler, "_unload_controlnet"
        ) as mock_unload, patch.object(
            StableDiffusionModelManager,
            "controlnet_is_loading",
            new_callable=PropertyMock,
        ) as mock_loading:
            # Test that it doesn't call _unload_controlnet when controlnet_is_loading is True
            mock_loading.return_value = True
            self.handler.unload_controlnet()
            mock_unload.assert_not_called()
            mock_unload.reset_mock()

            # Test that it calls _unload_controlnet when controlnet_is_loading is False
            mock_loading.return_value = False
            self.handler.unload_controlnet()
            mock_unload.assert_called_once()

    def test_load_scheduler(self):
        """Test the load_scheduler method"""
        with patch.object(self.handler, "_load_scheduler") as mock_load:
            scheduler_name = "MockScheduler"
            self.handler.load_scheduler(scheduler_name)
            mock_load.assert_called_once_with(scheduler_name)

    def test_reload(self):
        """Test the reload method"""
        with patch.object(self.handler, "unload") as mock_unload, patch.object(
            self.handler, "load"
        ) as mock_load, patch.object(
            self.handler, "_clear_cached_properties"
        ) as mock_clear:
            self.handler.reload()
            mock_clear.assert_called_once()
            mock_unload.assert_called_once()
            mock_load.assert_called_once()

    def test_load(self):
        """Test the load method"""
        # Case 1: Already loading or loaded
        with patch.object(
            StableDiffusionModelManager,
            "sd_is_loading",
            new_callable=PropertyMock,
        ) as mock_is_loading, patch.object(
            self.handler, "_load_safety_checker"
        ) as mock_load_safety:
            mock_is_loading.return_value = True
            self.handler.load()
            mock_load_safety.assert_not_called()

        # Case 2: Normal load case
        with patch.object(
            StableDiffusionModelManager,
            "sd_is_loading",
            new_callable=PropertyMock,
        ) as mock_is_loading, patch.object(
            StableDiffusionModelManager,
            "sd_is_loaded",
            new_callable=PropertyMock,
        ) as mock_is_loaded, patch.object(
            self.handler, "_load_safety_checker"
        ) as mock_load_safety, patch.object(
            self.handler, "_load_controlnet"
        ) as mock_load_controlnet, patch.object(
            self.handler, "_load_pipe"
        ) as mock_load_pipe, patch.object(
            self.handler, "_load_scheduler"
        ) as mock_load_scheduler, patch.object(
            self.handler, "_load_lora"
        ) as mock_load_lora, patch.object(
            self.handler, "_load_embeddings"
        ) as mock_load_embeddings, patch.object(
            self.handler, "_load_compel"
        ) as mock_load_compel, patch.object(
            self.handler, "_load_deep_cache"
        ) as mock_load_deep_cache, patch.object(
            self.handler, "_make_memory_efficient"
        ) as mock_memory_efficient, patch.object(
            self.handler, "_finalize_load_stable_diffusion"
        ) as mock_finalize:

            mock_is_loading.return_value = False
            mock_is_loaded.return_value = False

            self.handler.load()

            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADING
            )
            mock_load_safety.assert_called_once()
            mock_load_controlnet.assert_called_once()
            mock_load_pipe.assert_called_once()
            mock_load_scheduler.assert_called_once()
            mock_load_lora.assert_called_once()
            mock_load_embeddings.assert_called_once()
            mock_load_compel.assert_called_once()
            mock_load_deep_cache.assert_called_once()
            mock_memory_efficient.assert_called_once()
            mock_finalize.assert_called_once()

    def test_unload(self):
        """Test the unload method"""
        # Case 1: Already unloading or unloaded
        with patch.object(
            StableDiffusionModelManager,
            "sd_is_loading",
            new_callable=PropertyMock,
        ) as mock_is_loading, patch.object(
            self.handler, "_unload_safety_checker"
        ) as mock_unload_safety:
            mock_is_loading.return_value = True
            self.handler.unload()
            mock_unload_safety.assert_not_called()

        # Case 2: Normal unload case
        with patch.object(
            StableDiffusionModelManager,
            "sd_is_loading",
            new_callable=PropertyMock,
        ) as mock_is_loading, patch.object(
            StableDiffusionModelManager,
            "sd_is_unloaded",
            new_callable=PropertyMock,
        ) as mock_is_unloaded, patch.object(
            self.handler, "_unload_safety_checker"
        ) as mock_unload_safety, patch.object(
            self.handler, "_unload_scheduler"
        ) as mock_unload_scheduler, patch.object(
            self.handler, "_unload_controlnet"
        ) as mock_unload_controlnet, patch.object(
            self.handler, "_unload_loras"
        ) as mock_unload_loras, patch.object(
            self.handler, "_unload_emebeddings"
        ) as mock_unload_embeddings, patch.object(
            self.handler, "_unload_compel"
        ) as mock_unload_compel, patch.object(
            self.handler, "_unload_generator"
        ) as mock_unload_generator, patch.object(
            self.handler, "_unload_deep_cache"
        ) as mock_unload_deep_cache, patch.object(
            self.handler, "_unload_pipe"
        ) as mock_unload_pipe, patch.object(
            self.handler, "_clear_memory_efficient_settings"
        ) as mock_clear_settings, patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.clear_memory"
        ) as mock_clear_memory:

            mock_is_loading.return_value = False
            mock_is_unloaded.return_value = False
            self.handler._current_state = HandlerState.READY

            self.handler.unload()

            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADING
            )
            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.UNLOADED
            )
            mock_unload_safety.assert_called_once()
            mock_unload_scheduler.assert_called_once()
            mock_unload_controlnet.assert_called_once()
            mock_unload_loras.assert_called_once()
            mock_unload_embeddings.assert_called_once()
            mock_unload_compel.assert_called_once()
            mock_unload_generator.assert_called_once()
            mock_unload_deep_cache.assert_called_once()
            mock_unload_pipe.assert_called_once()
            mock_clear_settings.assert_called_once()
            mock_clear_memory.assert_called_once()

    def test_handle_generate_signal(self):
        """Test the handle_generate_signal method"""
        # Create mock image request
        mock_image_request = MagicMock()
        mock_image_request.scheduler = "TestScheduler"
        mock_image_request.callback = None

        # Create mock message
        mock_message = {"image_request": mock_image_request}

        with patch.object(
            self.handler, "_load_scheduler"
        ) as mock_load_scheduler, patch.object(
            self.handler, "load"
        ) as mock_load, patch.object(
            self.handler, "_clear_cached_properties"
        ) as mock_clear_cached, patch.object(
            self.handler, "_swap_pipeline"
        ) as mock_swap_pipeline, patch.object(
            self.handler, "_generate"
        ) as mock_generate, patch.object(
            self.handler, "emit_signal"
        ) as mock_emit, patch.object(
            self.handler, "handle_requested_action"
        ) as mock_handle_action, patch(
            "airunner.handlers.stablediffusion.stable_diffusion_model_manager.clear_memory"
        ) as mock_clear_memory:

            # Setup the current state and mock behavior
            self.handler._current_state = HandlerState.READY
            mock_generate.return_value = "Generated image"

            # Call the method
            self.handler.handle_generate_signal(mock_message)

            # Verify the calls
            self.assertEqual(self.handler.image_request, mock_image_request)
            mock_load_scheduler.assert_called_with("TestScheduler")
            mock_load.assert_called_once()
            mock_clear_cached.assert_called_once()
            mock_swap_pipeline.assert_called_once()
            mock_generate.assert_called_once()
            mock_emit.assert_called_with(
                SignalCode.ENGINE_RESPONSE_WORKER_RESPONSE_SIGNAL,
                {
                    "code": EngineResponseCode.IMAGE_GENERATED,
                    "message": "Generated image",
                },
            )
            self.assertEqual(self.handler._current_state, HandlerState.READY)
            mock_clear_memory.assert_called_once()
            mock_handle_action.assert_called_once()

    def test_reload_lora(self):
        """Test the reload_lora method"""
        # Setup the handler to simulate different model status conditions
        with patch.object(
            self.handler, "_unload_loras"
        ) as mock_unload, patch.object(
            self.handler, "_load_lora"
        ) as mock_load, patch.object(
            self.handler, "emit_signal"
        ) as mock_emit:

            # Case 1: SD is not loaded or generating, should not reload lora
            self.mock_model_status[ModelType.SD] = ModelStatus.FAILED
            self.handler._current_state = HandlerState.READY
            self.handler.reload_lora()
            mock_unload.assert_not_called()
            mock_load.assert_not_called()
            mock_emit.assert_not_called()

            # Case 2: SD is loaded and not generating, should reload lora
            self.mock_model_status[ModelType.SD] = ModelStatus.LOADED
            self.handler._current_state = HandlerState.READY

            self.handler.reload_lora()

            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADING
            )
            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADED
            )
            mock_unload.assert_called_once()
            mock_load.assert_called_once()
            mock_emit.assert_called_with(SignalCode.LORA_UPDATED_SIGNAL)

    def test_reload_embeddings(self):
        """Test the reload_embeddings method"""
        # Setup the handler to simulate different model status conditions
        with patch.object(
            self.handler, "_load_embeddings"
        ) as mock_load, patch.object(self.handler, "emit_signal") as mock_emit:

            # Case 1: SD is not loaded or generating, should not reload embeddings
            self.mock_model_status[ModelType.SD] = ModelStatus.FAILED
            self.handler._current_state = HandlerState.READY
            self.handler.reload_embeddings()
            mock_load.assert_not_called()
            mock_emit.assert_not_called()

            # Case 2: SD is loaded and not generating, should reload embeddings
            self.mock_model_status[ModelType.SD] = ModelStatus.LOADED
            self.handler._current_state = HandlerState.READY

            self.handler.reload_embeddings()

            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADING
            )
            self.handler.change_model_status.assert_any_call(
                ModelType.SD, ModelStatus.LOADED
            )
            mock_load.assert_called_once()
            mock_emit.assert_called_with(SignalCode.EMBEDDING_UPDATED_SIGNAL)

    def test_load_embeddings(self):
        """Test the load_embeddings method"""
        with patch.object(self.handler, "_load_embeddings") as mock_load:
            self.handler.load_embeddings()
            mock_load.assert_called_once()

    def test_interrupt_image_generation(self):
        """Test the interrupt_image_generation method"""
        # Test when current state is not generating
        self.handler._current_state = HandlerState.READY
        self.handler.do_interrupt_image_generation = False
        self.handler.interrupt_image_generation()
        self.assertFalse(self.handler.do_interrupt_image_generation)

        # Test when current state is generating
        self.handler._current_state = HandlerState.GENERATING
        self.handler.do_interrupt_image_generation = False
        self.handler.interrupt_image_generation()
        self.assertTrue(self.handler.do_interrupt_image_generation)

    def test_resize_image_static_method(self):
        """Test the _resize_image static method"""
        # Create a test image
        test_image = Image.new("RGB", (200, 100), color="red")

        # Test case: Image already fits within max dimensions
        result = StableDiffusionModelManager._resize_image(
            test_image, 300, 200
        )
        self.assertEqual(result.size, (200, 100))

        # Test case: Image needs to be resized (landscape orientation)
        result = StableDiffusionModelManager._resize_image(
            test_image, 100, 200
        )
        self.assertEqual(
            result.size[0], 100
        )  # Width should be exactly max_width
        self.assertLess(
            result.size[1], 100
        )  # Height should be proportionally smaller

        # Test case: Image is None
        result = StableDiffusionModelManager._resize_image(None, 100, 100)
        self.assertIsNone(result)


if __name__ == "__main__":
    unittest.main()



from airunner.vendor.nodegraphqt import (
    NodeGraph,
    PropertiesBinWidget,
)
from PySide6.QtCore import Qt


from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.gui.widgets.nodegraph.nodes.core.variable_getter_node import (
    VariableGetterNode,
)

# Define the MIME type constant
VARIABLE_MIME_TYPE = "application/x-airunner-variable"


class CustomNodeGraph(MediatorMixin, SettingsMixin, NodeGraph):
    """Custom NodeGraph class to handle specific events like variable drops."""

    widget_ref = None

    def __init__(self, parent=None):
        super(CustomNodeGraph, self).__init__(parent)
        self._prop_bin = PropertiesBinWidget(node_graph=self)
        self._prop_bin.setWindowFlags(Qt.WindowType.Tool)
        self.node_double_clicked.connect(self.display_prop_bin)
        self.port_connected.connect(self._on_port_connected)
        self.viewer().setAcceptDrops(True)

    def display_prop_bin(self, node):
        """
        function for displaying the properties bin when a node
        is double clicked
        """
        if not self._prop_bin.isVisible():
            self._prop_bin.show()

    def _on_port_connected(self, in_port, out_port):
        """
        Handle connections between ports by transferring values.
        This is particularly important for variable nodes to transfer their values.
        """
        # Check if the output port is from a VariableGetterNode
        if isinstance(out_port.node(), VariableGetterNode):
            var_node = out_port.node()
            # Get the variable value
            variable = None
            if (
                hasattr(self, "widget_ref")
                and self.widget_ref
                and hasattr(var_node, "variable_name")
            ):
                variable = self.widget_ref._find_variable_by_name(
                    var_node.variable_name
                )

            if variable:
                value = variable.get_value()
                target_node = in_port.node()
                port_name = in_port.name()

                # Set the value on the target node if it has the corresponding property
                if target_node.has_property(port_name):
                    target_node.set_property(port_name, value)
                    if hasattr(self, "widget_ref") and self.widget_ref:
                        self.logger.info(
                            f"Set property '{port_name}' on node '{target_node.name()}' to value '{value}' from variable '{var_node.variable_name}'"
                        )

                    # If the target node has an update method, call it
                    if hasattr(target_node, "update") and callable(
                        getattr(target_node, "update")
                    ):
                        target_node.update()

    # --- Drag and Drop Event Handling ---

    def dragEnterEvent(self, event):
        """Accept drag events if they contain our custom variable MIME type."""
        if event.mimeData().hasFormat(VARIABLE_MIME_TYPE):
            event.acceptProposedAction()
        else:
            # Important: Pass unhandled events to the base class
            super().dragEnterEvent(event)

    def dragMoveEvent(self, event):
        """Accept move events if they contain our custom variable MIME type."""
        if event.mimeData().hasFormat(VARIABLE_MIME_TYPE):
            event.acceptProposedAction()
        else:
            super().dragMoveEvent(event)

    def dropEvent(self, event):
        """Handle drop events for variables."""
        if event.mimeData().hasFormat(VARIABLE_MIME_TYPE):
            event.acceptProposedAction()

            variable_name = (
                event.mimeData().data(VARIABLE_MIME_TYPE).data().decode()
            )

            # Get the variable details (type) from the NodeGraphWidget
            if self.widget_ref and hasattr(
                self.widget_ref, "_find_variable_by_name"
            ):
                variable = self.widget_ref._find_variable_by_name(
                    variable_name
                )
                if variable:
                    # Create the VariableGetterNode at the drop position
                    # Convert scene position to graph position if necessary (might not be needed)
                    pos = self.viewer().mapToScene(event.pos())

                    # Use the correct registered identifier for VariableGetterNode
                    # Assuming default registration based on class path
                    node_identifier = "airunner.gui.widgets.nodegraph.nodes.variable_getter_node.VariableGetterNode"

                    try:
                        getter_node: VariableGetterNode = self.create_node(
                            node_identifier,
                            pos=[pos.x(), pos.y()],
                        )
                        if getter_node:
                            getter_node.set_variable(
                                variable.name, variable.var_type
                            )
                            self.logger.info(
                                f"Created Getter node for variable: {variable.name} at {pos.x()},{pos.y()}"
                            )
                        else:
                            self.logger.error(
                                f"Failed to create node with identifier: {node_identifier}"
                            )
                    except Exception as e:
                        self.logger.error(
                            f"Error creating VariableGetterNode: {e}",
                            exc_info=True,
                        )

                else:
                    if self.widget_ref:
                        self.logger.warning(
                            f"Dropped variable '{variable_name}' not found."
                        )
            else:
                print(
                    "Error: Cannot access variable list from graph (widget_ref missing or invalid)."
                )

        else:
            # Important: Pass unhandled events to the base class for default node dropping etc.
            super().dropEvent(event)



from typing import Dict
import json
import re
import time

from PySide6.QtCore import Signal, QRect, QThread, QObject, Slot, QSettings
from PySide6.QtWidgets import QApplication, QWidget

from airunner.data.models import ShortcutKeys, AIModels
from airunner.enums import (
    QualityEffects,
    SignalCode,
    GeneratorSection,
    ImagePreset,
    StableDiffusionVersion,
    ModelStatus,
    ModelType,
    LLMActionType,
)
from airunner.utils.application.mediator_mixin import MediatorMixin
from airunner.utils.application import random_seed
from airunner.gui.widgets.base_widget import BaseWidget
from airunner.gui.widgets.stablediffusion.templates.stablediffusion_generator_form_ui import (
    Ui_stablediffusion_generator_form,
)
from airunner.gui.widgets.stablediffusion.prompt_container_widget import (
    PromptContainerWidget,
)
from airunner.gui.windows.main.settings_mixin import SettingsMixin
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.handlers.stablediffusion.image_request import ImageRequest
from airunner.utils.widgets import load_splitter_settings


class SaveGeneratorSettingsWorker(
    MediatorMixin,
    SettingsMixin,
    QObject,
):
    def __init__(self, parent):
        super().__init__()
        self.parent = parent
        self.current_prompt_value = None
        self.current_negative_prompt_value = None
        self.current_secondary_prompt_value = None
        self.current_secondary_negative_prompt_value = None
        self.crops_coords_top_left_x = 0
        self.crops_coords_top_left_y = 0

    def run(self):
        do_update_settings = False
        while True:
            value = self.parent.ui.prompt.toPlainText()
            if value != self.current_prompt_value:
                self.current_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.negative_prompt.toPlainText()
            if value != self.current_negative_prompt_value:
                self.current_negative_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.secondary_prompt.toPlainText()
            if value != self.current_secondary_prompt_value:
                self.current_secondary_prompt_value = value
                do_update_settings = True

            value = self.parent.ui.secondary_negative_prompt.toPlainText()
            if value != self.current_secondary_negative_prompt_value:
                self.current_secondary_negative_prompt_value = value
                do_update_settings = True

            x = self.parent.ui.crops_coords_top_left_x.text()
            y = self.parent.ui.crops_coords_top_left_y.text()
            x = int(x) if x != "" else 0
            y = int(y) if y != "" else 0

            if self.crops_coords_top_left_x != x:
                self.crops_coords_top_left_x = x
                do_update_settings = True
            if self.crops_coords_top_left_y != y:
                self.crops_coords_top_left_y = y
                do_update_settings = True

            if do_update_settings:
                do_update_settings = False
                generator_settings = self.generator_settings
                generator_settings.prompt = self.current_prompt_value
                generator_settings.negative_prompt = (
                    self.current_negative_prompt_value
                )
                generator_settings.second_prompt = (
                    self.current_secondary_prompt_value
                )
                generator_settings.second_negative_prompt = (
                    self.current_secondary_negative_prompt_value
                )
                generator_settings.crops_coords_top_left = {
                    "x": self.crops_coords_top_left_x,
                    "y": self.crops_coords_top_left_y,
                }
                generator_settings.save()

            time.sleep(0.1)


class StableDiffusionGeneratorForm(BaseWidget):
    widget_class_ = Ui_stablediffusion_generator_form
    changed_signal = Signal(str, object)
    _prompt_containers: Dict[str, QWidget] = {}
    icons = [
        ("chevron-up", "generate_button"),
        ("x", "interrupt_button"),
        ("circle", "infinite_images_button"),
    ]

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.SD_GENERATE_IMAGE_SIGNAL: self.on_generate_image_signal,
            SignalCode.APPLICATION_STOP_SD_PROGRESS_BAR_SIGNAL: self.on_stop_image_generator_progress_bar_signal,
            SignalCode.SD_PROGRESS_SIGNAL: self.on_progress_signal,
            SignalCode.GENERATOR_FORM_UPDATE_VALUES_SIGNAL: self.set_form_values,
            SignalCode.LLM_IMAGE_PROMPT_GENERATED_SIGNAL: self.on_llm_image_prompt_generated_signal,
            SignalCode.GENERATE_IMAGE_FROM_IMAGE_SIGNAL: self.handle_generate_image_from_image,
            SignalCode.DO_GENERATE_IMAGE_FROM_IMAGE_SIGNAL: self.do_generate_image_from_image_signal_handler,
            SignalCode.SD_LOAD_PROMPT_SIGNAL: self.on_load_saved_stablediffuion_prompt_signal,
            SignalCode.BOT_MOOD_UPDATED: self.on_bot_mood_updated,
            SignalCode.KEYBOARD_SHORTCUTS_UPDATED: self.on_keyboard_shortcuts_updated,
            SignalCode.MODEL_STATUS_CHANGED_SIGNAL: self.on_model_status_changed_signal,
            SignalCode.CLEAR_PROMPTS: self.clear_prompts,
            SignalCode.WIDGET_ELEMENT_CHANGED: self.on_widget_element_changed,
            SignalCode.SD_ADDITIONAL_PROMPT_DELETE_SIGNAL: self.on_delete_prompt_clicked,
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed,
        }
        super().__init__(*args, **kwargs)
        self.seed_override = None
        self.parent = None
        self.initialized = False
        self.thread = QThread()
        self.worker = SaveGeneratorSettingsWorker(parent=self)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self._sd_version: str = self.generator_settings.version
        self._toggle_sdxl_form_elements()
        self.toggle_microconditioning(
            self.generator_settings.quality_effects
            == QualityEffects.CUSTOM.value
        )
        self.ui.quality_effects.blockSignals(True)
        self.ui.infinite_images_button.blockSignals(True)
        self.ui.quality_effects.clear()
        self.ui.quality_effects.addItems(
            [effect.value for effect in QualityEffects]
        )
        self.ui.quality_effects.setCurrentText(
            self.generator_settings.quality_effects
        )
        self.ui.infinite_images_button.setChecked(
            self.generator_settings.generate_infinite_images if self.generator_settings.generate_infinite_images is not None else False
        )
        self.ui.quality_effects.blockSignals(False)
        self.ui.infinite_images_button.blockSignals(False)

    @property
    def is_sd_xl_or_turbo(self) -> bool:
        return (
            self._sd_version == StableDiffusionVersion.SDXL1_0.value
            or self._sd_version == StableDiffusionVersion.SDXL_TURBO.value
        )

    @Slot()
    def on_generate_button_clicked(self):
        self.handle_generate_button_clicked()

    @Slot()
    def on_interrupt_button_clicked(self):
        self.api.art.canvas.interrupt_image_generation()

    def on_delete_prompt_clicked(self, data: Dict):
        prompt_id = data.get("prompt_id", None)
        if prompt_id is None:
            self.logger.error(f"Unable to delete prompt")
            return
        prompt_container = self._prompt_containers[prompt_id]
        self.ui.additional_prompts_container_layout.removeWidget(
            prompt_container
        )
        prompt_container.deleteLater()
        self._prompt_containers.pop(prompt_id)

        # Save the updated prompt containers after deletion
        self.save_prompt_containers_to_settings()

    def on_application_settings_changed(self, data: Dict):
        if data.get("setting_name") == "generator_settings":
            if data.get("column_name") in ("use_compel",):
                self._toggle_compel_form_elements(data.get("value", True))

    def on_widget_element_changed(self, data: Dict):
        self._toggle_compel_form_elements(self.generator_settings.use_compel)

        if data.get("element") in ("sd_version",):
            self._sd_version = data.get("version")
            self._toggle_sdxl_form_elements()

    def _toggle_compel_form_elements(self, value: bool):
        self.logger.info("Toggle compel form elements")
        # Iterate over all widgets in the layout and enable/disable them
        for i in range(self.ui.additional_prompts_container_layout.count()):
            widget = self.ui.additional_prompts_container_layout.itemAt(
                i
            ).widget()
            if widget:
                widget.show() if value else widget.hide()

    def _toggle_sdxl_form_elements(self):
        if self.is_sd_xl_or_turbo:
            self.ui.sdxl_settings_container.show()
            self.ui.secondary_prompt.show()
            self.ui.secondary_negative_prompt.show()
            self.ui.original_size_width.blockSignals(True)
            self.ui.original_size_height.blockSignals(True)
            self.ui.negative_original_size_width.blockSignals(True)
            self.ui.negative_original_size_height.blockSignals(True)
            self.ui.target_size_width.blockSignals(True)
            self.ui.target_size_height.blockSignals(True)
            self.ui.negative_target_size_width.blockSignals(True)
            self.ui.negative_target_size_height.blockSignals(True)
            self.ui.crops_coords_top_left_x.blockSignals(True)
            self.ui.crops_coords_top_left_y.blockSignals(True)
            self.ui.negative_crops_coord_top_left_x.blockSignals(True)
            self.ui.negative_crops_coord_top_left_y.blockSignals(True)
            self.ui.original_size_width.setText(
                str(
                    (self.generator_settings.original_size or {}).get(
                        "width", 0
                    )
                )
            )
            self.ui.original_size_height.setText(
                str(
                    (self.generator_settings.original_size or {}).get(
                        "height", 0
                    )
                )
            )
            self.ui.negative_original_size_width.setText(
                str(
                    (self.generator_settings.negative_original_size or {}).get(
                        "width", 0
                    )
                )
            )
            self.ui.negative_original_size_height.setText(
                str(
                    (self.generator_settings.negative_original_size or {}).get(
                        "height", 0
                    )
                )
            )
            self.ui.target_size_width.setText(
                str(
                    (self.generator_settings.target_size or {}).get("width", 0)
                )
            )
            self.ui.target_size_height.setText(
                str(
                    (self.generator_settings.target_size or {}).get(
                        "height", 0
                    )
                )
            )
            self.ui.negative_target_size_width.setText(
                str(
                    (self.generator_settings.negative_target_size or {}).get(
                        "width", 0
                    )
                )
            )
            self.ui.negative_target_size_height.setText(
                str(
                    (self.generator_settings.negative_target_size or {}).get(
                        "height", 0
                    )
                )
            )
            self.ui.crops_coords_top_left_x.setText(
                str(
                    (self.generator_settings.crops_coords_top_left or {}).get(
                        "x", 0
                    )
                )
            )
            self.ui.crops_coords_top_left_y.setText(
                str(
                    (self.generator_settings.crops_coords_top_left or {}).get(
                        "y", 0
                    )
                )
            )
            self.ui.negative_crops_coord_top_left_x.setText(
                str(
                    (
                        self.generator_settings.negative_crops_coords_top_left
                        or {}
                    ).get("x", 0)
                )
            )
            self.ui.negative_crops_coord_top_left_y.setText(
                str(
                    (
                        self.generator_settings.negative_crops_coords_top_left
                        or {}
                    ).get("y", 0)
                )
            )
            self.ui.original_size_width.blockSignals(False)
            self.ui.original_size_height.blockSignals(False)
            self.ui.negative_original_size_width.blockSignals(False)
            self.ui.negative_original_size_height.blockSignals(False)
            self.ui.target_size_width.blockSignals(False)
            self.ui.target_size_height.blockSignals(False)
            self.ui.negative_target_size_width.blockSignals(False)
            self.ui.negative_target_size_height.blockSignals(False)
            self.ui.crops_coords_top_left_x.blockSignals(False)
            self.ui.crops_coords_top_left_y.blockSignals(False)
            self.ui.negative_crops_coord_top_left_x.blockSignals(False)
            self.ui.negative_crops_coord_top_left_y.blockSignals(False)
        else:
            self.ui.sdxl_settings_container.hide()
            self.ui.secondary_prompt.hide()
            self.ui.secondary_negative_prompt.hide()

    @property
    def is_txt2img(self):
        return self.pipeline_action == GeneratorSection.TXT2IMG.value

    @property
    def is_outpaint(self):
        return self.pipeline_action == GeneratorSection.OUTPAINT.value

    @property
    def pipeline_action(self):
        return self.generator_settings.pipeline_action

    @property
    def generator_name(self):
        return self.application_settings.current_image_generator

    @property
    def seed(self):
        return self.generator_settings.seed

    @seed.setter
    def seed(self, val):
        self.update_generator_settings("seed", val)

    @property
    def active_rect(self):
        pos = self.active_grid_settings.pos
        rect = QRect(
            pos[0],
            pos[1],
            self.application_settings.working_width,
            self.application_settings.working_height,
        )
        drawing_pad_pos = self.drawing_pad_settings.pos
        rect.translate(-drawing_pad_pos[0], -drawing_pad_pos[1])

        return rect

    def on_keyboard_shortcuts_updated(self):
        self._set_keyboard_shortcuts()

    def on_bot_mood_updated(self, data):
        pass

    def on_generate_image_signal(self, _data):
        self.handle_generate_button_clicked()

    def on_stop_image_generator_progress_bar_signal(self, data: Dict):
        self.stop_progress_bar(data.get("do_clear", False))

    def on_progress_signal(self, message):
        self.handle_progress_bar(message)

    ##########################################################################
    # LLM Generated Image handlers
    ##########################################################################
    def on_llm_image_prompt_generated_signal(self, data):
        """
        This slot is called after an LLM has generated the prompts for an image.
        It sets the prompts in the generator form UI and continues the image generation process.
        """
        # Unload non-Stable Diffusion models
        self.api.art.unload_non_sd(callback=self.unload_llm_callback)
        # Set the prompts in the generator form UI
        data = data["message"]
        self.update_application_settings("working_width", data["width"])
        self.update_application_settings("working_height", data["height"])
        self.update_generator_settings("image_preset", data.get("type", ""))
        prompt = data.get("prompt", None)
        secondary_prompt = data.get("secondary_prompt", None)
        self.ui.prompt.setPlainText(prompt)
        self.ui.secondary_prompt.setPlainText(secondary_prompt)

    def unload_llm_callback(self, _data: dict = None):
        """
        Callback function to be called after the LLM has been unloaded.
        """
        if not self.application_settings.sd_enabled:
            # If SD is not enabled, enable it and then emit a signal to generate the image
            # The callback function is handled by the signal handler for the SD_LOAD_SIGNAL.
            # The finalize function is a callback which is called after the image has been generated.
            self.logger.info(
                "Stable Diffusion is not enabled, enabling it now."
            )
            self.api.art.toggle_sd(
                enabled=True,
                callback=self.handle_generate_button_clicked,
                finalize=self.finalize_image_generated_by_llm,
            )
        else:
            # If SD is already enabled, emit a signal to generate the image.
            # The finalize function is a callback which is called after the image has been generated.
            self.logger.info(
                "Stable Diffusion is already enabled, generating the image."
            )
            self.handle_generate_button_clicked(
                dict(
                    enabled=True, finalize=self.finalize_image_generated_by_llm
                )
            )

    def finalize_image_generated_by_llm(self, _data):
        """
        Callback function to be called after the image has been generated.
        """
        self.api.art.toggle_sd(
            enabled=False,
            callback=lambda _d: self.load_non_sd(
                callback=lambda _d: self.api.llm.send_llm_text_streamed_signal(
                    LLMResponse(
                        message="Your image has been generated",
                        is_first_message=True,
                        is_end_of_message=True,
                        name=self.chatbot.botname,
                        action=LLMActionType.GENERATE_IMAGE,
                    )
                )
            ),
        )

    ##########################################################################
    # End LLM Generated Image handlers
    ##########################################################################

    @Slot()
    def handle_add_prompt_clicked(self):
        additional_prompts_container_layout = (
            self.ui.additional_prompts_container_layout
        )
        prompt_container = PromptContainerWidget()
        prompt_container.prompt_id = len(self._prompt_containers.keys())
        additional_prompts_container_layout.addWidget(prompt_container)

        # store prompt container in self._prompt_containers
        self._prompt_containers[prompt_container.prompt_id] = prompt_container

        # Save the updated prompt containers
        self.save_prompt_containers_to_settings()

    def handle_generate_image_from_image(self, image):
        pass

    def on_load_saved_stablediffuion_prompt_signal(self, data: dict):
        saved_prompt = data.get("saved_prompt")
        self.ui.prompt.blockSignals(True)
        self.ui.negative_prompt.blockSignals(True)
        self.ui.secondary_prompt.blockSignals(True)
        self.ui.secondary_negative_prompt.blockSignals(True)
        self.ui.prompt.setPlainText(saved_prompt.prompt)
        self.ui.negative_prompt.setPlainText(saved_prompt.negative_prompt)
        self.ui.secondary_prompt.setPlainText(saved_prompt.secondary_prompt)
        self.ui.secondary_negative_prompt.setPlainText(
            saved_prompt.secondary_negative_prompt
        )
        self.ui.prompt.blockSignals(False)
        self.ui.negative_prompt.blockSignals(False)
        self.ui.secondary_prompt.blockSignals(False)
        self.ui.secondary_negative_prompt.blockSignals(False)

    def handle_image_presets_changed(self, val):
        self.update_generator_settings("image_preset", val)

    def do_generate_image_from_image_signal_handler(self, _data):
        self.do_generate()

    def do_generate(self, data=None):
        data = data or {}

        callback = data.get("finalize", None)

        # Update data with additional prompt data from self._prompt_containers
        additional_prompts = [
            {
                "prompt": container.get_prompt(),
                "prompt_secondary": container.get_prompt_secondary(),
            }
            for _prompt_id, container in self._prompt_containers.items()
        ]

        model_path = ""
        model_id = self.generator_settings.model
        if model_id is not None:
            aimodel = AIModels.objects.get(model_id)
            if aimodel is not None:
                model_path = model_path

        if model_path == "":
            if self.generator_settings.model is not None:
                aimodel = AIModels.objects.get(self.generator_settings.model)
            else:
                aimodel = AIModels.objects.first()

            if aimodel is not None:
                model_path = aimodel.path
                self.update_generator_settings("model", aimodel.id)

        image_request = ImageRequest(
            prompt=data.get("prompt", self.ui.prompt.toPlainText()),
            negative_prompt=data.get(
                "negative_prompt", self.ui.negative_prompt.toPlainText()
            ),
            second_prompt=data.get(
                "second_prompt", self.ui.secondary_prompt.toPlainText()
            ),
            second_negative_prompt=data.get(
                "second_negative_prompt",
                self.ui.secondary_negative_prompt.toPlainText(),
            ),
            crops_coords_top_left=self.generator_settings.crops_coords_top_left,
            negative_crops_coords_top_left=self.generator_settings.negative_crops_coords_top_left,
            pipeline_action=self.generator_settings.pipeline_action,
            generator_name=self.generator_name,
            random_seed=self.generator_settings.random_seed,
            model_path=model_path,
            scheduler=self.generator_settings.scheduler,
            version=self.generator_settings.version,
            use_compel=self.generator_settings.use_compel,
            steps=self.generator_settings.steps,
            ddim_eta=self.generator_settings.ddim_eta,
            scale=self.generator_settings.scale / 100,
            seed=self.seed,
            strength=self.generator_settings.strength / 100,
            n_samples=self.generator_settings.n_samples,
            images_per_batch=self.generator_settings.images_per_batch,
            generate_infinite_images=self.generator_settings.generate_infinite_images,
            clip_skip=self.generator_settings.clip_skip,
            width=self.application_settings.working_width,
            height=self.application_settings.working_height,
            target_size=self.generator_settings.target_size,
            original_size=self.generator_settings.original_size,
            negative_target_size=self.generator_settings.negative_target_size,
            negative_original_size=self.generator_settings.negative_original_size,
            lora_scale=self.generator_settings.lora_scale,
            additional_prompts=additional_prompts,
            callback=callback,
            image_preset=ImagePreset(self.generator_settings.image_preset),
            quality_effects=(
                QualityEffects(self.generator_settings.quality_effects)
                if self.generator_settings.quality_effects != ""
                and self.generator_settings.quality_effects is not None
                else QualityEffects.STANDARD
            ),
        )

        self.api.art.send_request(image_request=image_request)

    def action_clicked_button_save_prompts(self):
        self.api.art.save_prompt(
            prompt=self.ui.prompt.toPlainText(),
            negative_prompt=self.ui.negative_prompt.toPlainText(),
            secondary_prompt=self.ui.secondary_prompt.toPlainText(),
            secondary_negative_prompt=self.ui.secondary_negative_prompt.toPlainText(),
        )

    def handle_prompt_changed(self):
        pass

    def handle_negative_prompt_changed(self):
        pass

    def handle_second_prompt_changed(self):
        pass

    def handle_second_negative_prompt_changed(self):
        pass

    def handle_generate_button_clicked(self, data=None):
        self.start_progress_bar()
        self.generate(data)

    @Slot(bool)
    def on_infinite_images_button_toggled(self, val: bool):
        self.update_generator_settings("generate_infinite_images", val)

    @Slot(str)
    def on_original_size_width_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        original_size = self.generator_settings.original_size
        original_size = original_size or {}
        original_size["width"] = int(val)
        self.update_generator_settings("original_size", original_size)

    @Slot(str)
    def on_original_size_height_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        original_size = self.generator_settings.original_size
        original_size = original_size or {}
        original_size["height"] = int(val)
        self.update_generator_settings("original_size", original_size)

    @Slot(str)
    def on_negative_original_size_width_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_original_size = self.generator_settings.negative_original_size
        negative_original_size = negative_original_size or {}
        negative_original_size["width"] = int(val)
        self.update_generator_settings(
            "negative_original_size", negative_original_size
        )

    @Slot(str)
    def on_negative_original_size_height_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_original_size = self.generator_settings.negative_original_size
        negative_original_size = negative_original_size or {}
        negative_original_size["height"] = int(val)
        self.update_generator_settings(
            "negative_original_size", negative_original_size
        )

    @Slot(str)
    def on_target_size_width_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        target_size = self.generator_settings.target_size
        target_size = target_size or {}
        target_size["width"] = int(val)
        self.update_generator_settings("target_size", target_size)

    @Slot(str)
    def on_target_size_height_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        target_size = self.generator_settings.target_size
        target_size = target_size or {}
        target_size["height"] = int(val)
        self.update_generator_settings("target_size", target_size)

    @Slot(str)
    def on_negative_target_size_width_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_target_size = self.generator_settings.negative_target_size
        negative_target_size = negative_target_size or {}
        negative_target_size["width"] = int(val)
        self.update_generator_settings(
            "negative_target_size", negative_target_size
        )

    @Slot(str)
    def on_negative_target_size_height_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_target_size = self.generator_settings.negative_target_size
        negative_target_size = negative_target_size or {}
        negative_target_size["height"] = int(val)
        self.update_generator_settings(
            "negative_target_size", negative_target_size
        )

    @Slot(str)
    def on_crops_coords_top_left_x_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        crops_coords_top_left = self.generator_settings.crops_coords_top_left
        crops_coords_top_left = crops_coords_top_left or {}
        crops_coords_top_left["x"] = int(val)
        self.update_generator_settings(
            "crops_coords_top_left", crops_coords_top_left
        )

    @Slot(str)
    def on_crops_coords_top_left_y_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        crops_coords_top_left = self.generator_settings.crops_coords_top_left
        crops_coords_top_left = crops_coords_top_left or {}
        crops_coords_top_left["y"] = int(val)
        self.update_generator_settings(
            "crops_coords_top_left", crops_coords_top_left
        )

    @Slot(str)
    def on_negative_crops_coord_top_left_x_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_crops_coords_top_left = (
            self.generator_settings.negative_crops_coords_top_left
        )
        negative_crops_coords_top_left = negative_crops_coords_top_left or {}
        negative_crops_coords_top_left["x"] = int(val)
        self.update_generator_settings(
            "negative_crops_coords_top_left", negative_crops_coords_top_left
        )

    @Slot(str)
    def on_negative_crops_coord_top_left_y_textChanged(self, val: str):
        val = 0 if val == "" or val is None else val
        negative_crops_coords_top_left = (
            self.generator_settings.negative_crops_coords_top_left
        )
        negative_crops_coords_top_left = negative_crops_coords_top_left or {}
        negative_crops_coords_top_left["y"] = int(val)
        self.update_generator_settings(
            "negative_crops_coords_top_left", negative_crops_coords_top_left
        )

    def generate(self, data=None):
        if self.generator_settings.random_seed:
            self.seed = random_seed()
        self.do_generate(data)
        self.seed_override = None

    def do_generate_image(self):
        time.sleep(0.1)
        self.do_generate()

    def extract_json_from_message(self, message):
        # Regular expression to find the JSON block
        json_pattern = re.compile(r".*`json\s*({.*?})\s*`.*", re.DOTALL)
        match = json_pattern.search(message)

        if match:
            json_block = match.group(1)
            try:
                # Convert the JSON block to a dictionary
                json_dict = json.loads(json_block)
                return json_dict
            except json.JSONDecodeError as e:
                self.logger.error(f"Error decoding JSON block: {e}")
                return {}
        else:
            self.logger.error("No JSON block found in message")
            return {}

    def get_memory_options(self):
        return {
            "use_last_channels": self.memory_settings.use_last_channels,
            "use_enable_sequential_cpu_offload": self.memory_settings.use_enable_sequential_cpu_offload,
            "enable_model_cpu_offload": self.memory_settings.enable_model_cpu_offload,
            "use_attention_slicing": self.memory_settings.use_attention_slicing,
            "use_tf32": self.memory_settings.use_tf32,
            "use_cudnn_benchmark": self.memory_settings.use_cudnn_benchmark,
            "use_enable_vae_slicing": self.memory_settings.use_enable_vae_slicing,
            "use_accelerated_transformers": self.memory_settings.use_accelerated_transformers,
            "use_torch_compile": self.memory_settings.use_torch_compile,
            "use_tiled_vae": self.memory_settings.use_tiled_vae,
            "use_tome_sd": self.memory_settings.use_tome_sd,
            "tome_sd_ratio": self.memory_settings.tome_sd_ratio,
        }

    def handle_quality_effects_changed(self, val):
        self.update_generator_settings("quality_effects", val)
        self.toggle_microconditioning(val == QualityEffects.CUSTOM.value)

    def toggle_microconditioning(self, enabled: bool):
        self.ui.original_size_width.setEnabled(enabled)
        self.ui.original_size_height.setEnabled(enabled)
        self.ui.negative_original_size_width.setEnabled(enabled)
        self.ui.negative_original_size_height.setEnabled(enabled)
        self.ui.target_size_width.setEnabled(enabled)
        self.ui.target_size_height.setEnabled(enabled)
        self.ui.negative_target_size_width.setEnabled(enabled)
        self.ui.negative_target_size_height.setEnabled(enabled)

    def handle_progress_bar(self, message):
        step = message.get("step")
        total = message.get("total")
        if step == total:
            self.stop_progress_bar()
            return

        if step == 0 and total == 0:
            current = 0
        else:
            try:
                current = step / total
            except ZeroDivisionError:
                current = 0
        self.set_progress_bar_value(int(current * 100))

    def set_progress_bar_value(self, value):
        progressbar = self.ui.progress_bar
        if not progressbar:
            return
        if progressbar.maximum() == 0:
            progressbar.setRange(0, 100)
        progressbar.setValue(value)
        QApplication.processEvents()

    def start_progress_bar(self):
        self.ui.progress_bar.setFormat("Generating %p%")
        self.ui.progress_bar.setRange(0, 0)
        self.ui.progress_bar.show()

    def on_model_status_changed_signal(self, data):
        if data["model"] is ModelType.SD:
            if data["status"] is not ModelStatus.LOADING:
                self.stop_progress_bar(do_clear=True)
                self.ui.generate_button.setEnabled(True)
                self.ui.interrupt_button.setEnabled(True)
            else:
                self.start_progress_bar()
                self.ui.generate_button.setEnabled(False)
                self.ui.interrupt_button.setEnabled(False)

    def showEvent(self, event):
        super().showEvent(event)
        self.set_form_values()
        self.initialized = True
        self.thread.start()

        load_splitter_settings(self.ui, ["generator_form_splitter"])

        # Restore prompt containers when widget is shown
        self.restore_prompt_containers_from_settings()

    def hideEvent(self, event):
        """When widget is hidden, save prompt containers."""
        super().hideEvent(event)
        self.save_prompt_containers_to_settings()

    def closeEvent(self, event):
        """When widget is closed, save prompt containers."""
        self.save_prompt_containers_to_settings()
        super().closeEvent(event)

    def set_form_values(self, _data=None):
        self.ui.prompt.blockSignals(True)
        self.ui.negative_prompt.blockSignals(True)
        self.ui.secondary_prompt.blockSignals(True)
        self.ui.secondary_negative_prompt.blockSignals(True)
        self.ui.image_presets.blockSignals(True)
        self.ui.quality_effects.blockSignals(True)

        self.ui.prompt.setPlainText(self.generator_settings.prompt)
        self.ui.negative_prompt.setPlainText(
            self.generator_settings.negative_prompt
        )
        self.ui.secondary_prompt.setPlainText(
            self.generator_settings.second_prompt
        )
        self.ui.secondary_negative_prompt.setPlainText(
            self.generator_settings.second_negative_prompt
        )

        image_presets = [preset.value for preset in ImagePreset]
        self.ui.image_presets.addItems(image_presets)
        self.ui.image_presets.setCurrentIndex(
            self.ui.image_presets.findText(
                self.generator_settings.image_preset
            )
        )

        self.ui.quality_effects.setCurrentText(
            self.generator_settings.quality_effects
        )

        self.ui.prompt.blockSignals(False)
        self.ui.negative_prompt.blockSignals(False)
        self.ui.secondary_prompt.blockSignals(False)
        self.ui.secondary_negative_prompt.blockSignals(False)
        self.ui.image_presets.blockSignals(False)
        self.ui.quality_effects.blockSignals(False)

    def clear_prompts(self):
        self.ui.prompt.setPlainText("")
        self.ui.negative_prompt.setPlainText("")
        self.ui.secondary_prompt.setPlainText("")
        self.ui.secondary_negative_prompt.setPlainText("")

    def stop_progress_bar(self, do_clear=False):
        progressbar = self.ui.progress_bar
        if not progressbar:
            return
        progressbar.setRange(0, 100)
        if do_clear:
            progressbar.setValue(0)
            progressbar.setFormat("")
        else:
            progressbar.setValue(100)
            progressbar.setFormat("Complete")

    def _set_keyboard_shortcuts(self):
        generate_image_key = ShortcutKeys.objects.filter_by_first(
            display_name="Generate Image"
        )
        interrupt_key = ShortcutKeys.objects.filter_by_first(
            display_name="Interrupt"
        )
        if generate_image_key:
            self.ui.generate_button.setShortcut(generate_image_key.key)
            self.ui.generate_button.setToolTip(
                f"{generate_image_key.display_name} ({generate_image_key.text})"
            )
        if interrupt_key:
            self.ui.interrupt_button.setShortcut(interrupt_key.key)
            self.ui.interrupt_button.setToolTip(
                f"{interrupt_key.display_name} ({interrupt_key.text})"
            )

    def save_prompt_containers_to_settings(self):
        """Save all additional prompt containers to QSettings."""
        if not self.initialized:
            return

        settings = QSettings()
        settings.beginGroup("sd_additional_prompts")

        # Clear existing settings first
        settings.remove("")

        # Save the number of containers
        settings.setValue("count", len(self._prompt_containers))

        # Save each container's data
        for i, (prompt_id, container) in enumerate(
            self._prompt_containers.items()
        ):
            settings.setValue(f"prompt_{i}_id", prompt_id)
            settings.setValue(f"prompt_{i}_text", container.get_prompt())
            settings.setValue(
                f"prompt_{i}_text_secondary", container.get_prompt_secondary()
            )

        settings.endGroup()
        settings.sync()

    def restore_prompt_containers_from_settings(self):
        """Restore additional prompt containers from QSettings."""
        # Clear existing containers first
        for container in list(self._prompt_containers.values()):
            self.ui.additional_prompts_container_layout.removeWidget(container)
            container.deleteLater()
        self._prompt_containers.clear()

        settings = QSettings()
        settings.beginGroup("sd_additional_prompts")

        count = settings.value("count", 0, type=int)

        for i in range(count):
            prompt_id = settings.value(f"prompt_{i}_id", i, type=int)
            prompt_text = settings.value(f"prompt_{i}_text", "", type=str)
            prompt_text_secondary = settings.value(
                f"prompt_{i}_text_secondary", "", type=str
            )

            # Create and add the container
            prompt_container = PromptContainerWidget()
            prompt_container.prompt_id = prompt_id
            prompt_container.set_prompt(prompt_text)
            prompt_container.set_prompt_secondary(prompt_text_secondary)

            self.ui.additional_prompts_container_layout.addWidget(
                prompt_container
            )
            self._prompt_containers[prompt_id] = prompt_container

        settings.endGroup()
